# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import warnings
from collections import namedtuple
from typing import List, Union

import emcee
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_is_fitted

from eensight.methods.features import SafeOrdinalEncoder
from eensight.utils import as_list, get_categorical_cols

from .common import train_boost

###############################################################################################
# Utilities
###############################################################################################


def _create_lagged(X, lags):
    added_features = []
    for fea, lag_values in lags.items():
        if fea not in X.columns:
            continue
        for lag in as_list(lag_values):
            added_features.append(X[fea].shift(lag).to_frame(f"L{lag}.{fea}"))
    return pd.concat([X, *added_features], axis=1)


def _create_calendar(X):
    result = pd.DataFrame(0, columns=["hour", "day", "week"], index=X.index)
    result["hour"] = X.index.hour  # hour of day
    result["day"] = X.index.dayofweek  # day of week
    result["week"] = X.index.isocalendar().week  # week of year
    return result


def _log_prior(theta, y, priors):
    u_value, T_sp_coo, T_sp_hea, T_sb_coo, T_sb_hea, eta, ach, alpha, beta = theta

    u_value_min, u_value_max = priors.get("u_value", (0.5, 5))
    T_sp_coo_min, T_sp_coo_max = priors.get("T_sp_coo", (24, 27))
    T_sp_hea_min, T_sp_hea_max = priors.get("T_sp_hea", (20, 24))
    T_sb_coo_min, T_sb_coo_max = priors.get("T_sb_coo", (24, 35))
    T_sb_hea_min, T_sb_hea_max = priors.get("T_sb_hea", (5, 20))
    eta_min, eta_max = priors.get("eta", (0, 1))
    ach_min, ach_max = priors.get("ach", (0.3, 5))
    alpha_min, alpha_max = priors.get("alpha", (0, y.max() - y.min()))
    beta_min, beta_max = priors.get("beta", (0, y.min()))

    if (
        (u_value_min <= u_value <= u_value_max)
        and (T_sp_coo_min <= T_sp_coo <= T_sp_coo_max)
        and (T_sp_hea_min <= T_sp_hea <= T_sp_hea_max)
        and (T_sb_coo_min <= T_sb_coo <= T_sb_coo_max)
        and (T_sb_hea_min <= T_sb_hea <= T_sb_hea_max)
        and (eta_min <= eta <= eta_max)
        and (ach_min <= ach <= ach_max)
        and (alpha_min <= alpha <= alpha_max)
        and (beta_min <= beta <= beta_max)
    ):
        return 0.0

    return -np.inf


def _forward_model(theta, temperature, activity, A, V, cp, rho):
    u_value, T_sp_coo, T_sp_hea, T_sb_coo, T_sb_hea, eta, ach, alpha, beta = theta

    # air conditioning load when activity is positive
    cond_load_on = (
        A
        * u_value
        * np.maximum(temperature - T_sp_coo, T_sp_hea - temperature)
        .clip(lower=0)
        .pipe(lambda x: x.mask(activity <= 1e-05, 0))
    )

    # ventilation load
    ven_load = (
        (1 / 3600)
        * (1 - eta)
        * cp
        * rho
        * ach
        * V
        * np.maximum(temperature - T_sp_coo, T_sp_hea - temperature)
        .clip(lower=0)
        .pipe(lambda x: x.mask(activity <= 1e-05, 0))
    )

    # air conditioning load when activity is zero
    cond_load_off = (
        A
        * u_value
        * np.maximum(temperature - T_sb_coo, T_sb_hea - temperature)
        .clip(lower=0)
        .pipe(lambda x: x.mask(activity > 1e-05, 0))
    )

    # plug load
    plug_load = alpha * activity + beta

    return cond_load_on, ven_load, cond_load_off, plug_load


def _log_likelihood(theta, temperature, activity, y, A, V, cp, rho):
    _, T_sp_coo, T_sp_hea, T_sb_coo, T_sb_hea, eta, _, _, _ = theta

    if (
        (T_sp_coo < T_sp_hea)
        or (T_sb_coo < T_sb_hea)
        or (T_sp_coo > T_sb_coo)
        or (T_sp_hea < T_sb_hea)
        or (eta > 1)
        or (eta < 0)
    ):
        return -np.inf

    cond_load_on, ven_load, cond_load_off, plug_load = _forward_model(
        theta, temperature, activity, A, V, cp, rho
    )
    total_load = cond_load_on + ven_load + cond_load_off + plug_load

    return -0.5 * np.sum((y - total_load) ** 2)


def _log_probability(theta, temperature, activity, y, A, V, cp, rho, priors):
    lp = _log_prior(theta, y, priors)

    if not np.isfinite(lp):
        return -np.inf

    return lp + _log_likelihood(theta, temperature, activity, y, A, V, cp, rho)


def _run_mcmc(temperature, activity, y, *, A, V, cp, rho, run_params):
    ndim = 9  # number of parameters to estimate
    nwalkers = run_params["nwalkers"]
    priors = run_params["priors"]
    n_samples = run_params["n_samples"]
    verbose = run_params["verbose"]

    sampler = emcee.EnsembleSampler(
        nwalkers,
        ndim,
        _log_probability,
        args=(temperature, activity, y, A, V, cp, rho, priors),
    )

    u_value_min, u_value_max = priors.get("u_value", (0.5, 5))
    T_sp_coo_min, T_sp_coo_max = priors.get("T_sp_coo", (24, 27))
    T_sp_hea_min, T_sp_hea_max = priors.get("T_sp_hea", (20, 24))
    T_sb_coo_min, T_sb_coo_max = priors.get("T_sb_coo", (30, 35))
    T_sb_hea_min, T_sb_hea_max = priors.get("T_sb_hea", (5, 10))
    eta_min, eta_max = priors.get("eta", (0, 1))
    ach_min, ach_max = priors.get("ach", (0.3, 5))
    alpha_min, alpha_max = priors.get("alpha", (0, y.max() - y.min()))
    beta_min, beta_max = priors.get("beta", (0, y.min()))

    # initial values
    pos = np.array(
        [
            (u_value_min + u_value_max) / 2,  # u_value
            (T_sp_coo_min + T_sp_coo_max) / 2,  # T_sp_coo
            (T_sp_hea_min + T_sp_hea_max) / 2,  # T_sp_hea
            (T_sb_coo_min + T_sb_coo_max) / 2,  # T_sb_coo
            (T_sb_hea_min + T_sb_hea_max) / 2,  # T_sb_hea
            (eta_min + eta_max) / 2,  # eta
            (ach_min + ach_max) / 2,  # ach
            (alpha_min + alpha_max) / 2,  # alpha
            (beta_min + beta_max) / 2,  # beta
        ]
    )
    pos = np.tile(pos, (nwalkers, 1)) + np.random.randn(nwalkers, ndim)

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        sampler.run_mcmc(pos, n_samples, progress=verbose)

    chain = sampler.get_chain(discard=100, thin=15, flat=True)
    return chain


###############################################################################################
# Predictors
###############################################################################################


class UsagePredictor(RegressorMixin, BaseEstimator):
    """Predict energy consumption using Gradient Boosting regression.

    Args:
        lags (dict, optional): Dictionary with keys that correspond to feature names,
            and values that are lists containing the time lags of that feature to add
            as additional regressors. Defaults to None.
        cat_features (str or list of str, optional): The names of the categorical features
            in the input data. Defaults to None.
        validation_size (float, optional): The size of the validation dataset (as
            percentage of all training data) to use for identifying optimal number
            of iterations for the Gradient Boosting model. Defaults to 0.2.
        skip_calendar (bool, optional): If True, the model will not add calendar
            features to the input data. Defaults to False.
    """

    def __init__(
        self,
        lags: dict = None,
        cat_features: Union[str, List[str]] = None,
        validation_size: float = 0.2,
        skip_calendar: bool = False,
    ):
        self.lags = lags
        self.cat_features = cat_features
        self.validation_size = validation_size
        self.skip_calendar = skip_calendar

    def fit(self, X: pd.DataFrame, y: pd.DataFrame):
        """Fit the estimator with the available data.

        Args:
            X (pandas.DataFrame): Feature data.
            y (pandas.DataFrame): Target data.

        Returns:
            UsagePredictor: The fitted predictor.
        """

        self.has_neg_values_ = np.sum(y < 0).item() > 0
        cat_features = as_list(self.cat_features)

        if cat_features:
            self.cat_features_ = [name for name in cat_features if name in X.columns]
        else:
            self.cat_features_ = [
                name for name in get_categorical_cols(X, int_is_categorical=False)
            ]

        # Encode categorical features
        if self.cat_features_:
            self.ordinal_enc_ = SafeOrdinalEncoder(
                features=self.cat_features_, remainder="passthrough"
            )
            X = self.ordinal_enc_.fit_transform(X)
            X[self.cat_features_] = X[self.cat_features_].astype("category")

        if self.lags:
            X = _create_lagged(X, self.lags)

        if not self.skip_calendar:
            X = pd.concat([X, _create_calendar(X)], axis=1)

        model, scores = train_boost(
            X,
            y,
            cat_features=self.cat_features_,
            validation_size=self.validation_size,
            return_scores=True,
        )
        self.model_ = model
        self.scores_ = scores
        self.features_ = X.columns
        self.fitted_ = True
        return self

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        """Predict energy consumption given input data.

        Args:
            X (pandas.DataFrame): Input data.

        Returns:
            pandas.Series: The prediction.
        """
        check_is_fitted(self, "fitted_")

        # Encode categorical features
        if self.cat_features_:
            X = self.ordinal_enc_.transform(X)
            X[self.cat_features_] = X[self.cat_features_].astype("category")

        if self.lags:
            X = _create_lagged(X, self.lags)

        if not self.skip_calendar:
            X = pd.concat([X, _create_calendar(X)], axis=1)

        if not X.columns.equals(self.features_):
            raise ValueError(
                f"Columns to predict {X.columns} differ from "
                f"ones from training {self.features_}"
            )

        y_pred = pd.Series(self.model_.predict(X), index=X.index)

        if not self.has_neg_values_:
            y_pred = y_pred.clip(lower=0)

        return y_pred


def explain_predicition(
    X: pd.DataFrame,
    y: pd.Series,
    A: float,
    V: float,
    nwalkers: int = 100,
    n_samples: int = 1000,
    priors: dict = {},
    verbose: bool = False,
):
    """Explain a prediction by estimating first-principle parameters of the underlying building.

    Args:
        X (pandas.DataFrame): Feature data.
        y (pandas.Series): The prediction time series.
        A (float): Area of building exposed to outdoors (m.2).
        V (float): Air volume of building (m.3)
        nwalkers (int, optional): The number of walkers in the ensemble. Defaults to 100.
        n_samples (int, optional): The number of samples to collect. Defaults to 1000.
        priors (dict, optional): The priors for the parameters to be estimated. It should be
            a dictionary with keys: "u_value", "T_sp_coo", "T_sp_hea", "T_sb_coo", "T_sb_hea",
            "eta", "ach", "alpha", "beta", and values that are tuples of minimum and maximum
            expected values. Defaults to {}.
        verbose (bool, optional): If True, a progress bar will be visible. Defaults to False.

    Returns:
        namedtuple: A namedtuple with keys "cond_load_on", "ven_load", "cond_load_off", "plug_load".
    """

    run_params = {}

    run_params["nwalkers"] = nwalkers
    run_params["n_samples"] = n_samples
    run_params["priors"] = priors
    run_params["verbose"] = verbose

    temperature = X["temperature"]
    activity = X["activity"]

    samples = _run_mcmc(
        temperature, activity, y, A=A, V=V, cp=1.005, rho=1.2, run_params=run_params
    )
    theta = samples.mean(axis=0)

    cond_load_on, ven_load, cond_load_off, plug_load = _forward_model(
        theta, temperature, activity, A, V, 1.005, 1.2
    )

    Bunch = namedtuple(
        "Bunch", ["cond_load_on", "ven_load", "cond_load_off", "plug_load"]
    )
    return Bunch(
        cond_load_on=cond_load_on,
        ven_load=ven_load,
        cond_load_off=cond_load_off,
        plug_load=plug_load,
    )



from pathlib import Path

from kedro.framework.project import configure_project
from kedro.framework.startup import (
    ProjectMetadata,
    _add_src_to_path,
    _get_project_metadata,
)


def bootstrap_project(project_path: Path) -> ProjectMetadata:
    """Run setup required at the beginning of the workflow
    when running in project mode, and return project metadata.
    """
    metadata = _get_project_metadata(project_path)
    metadata = ProjectMetadata(
        config_file=metadata.config_file,
        package_name=metadata.package_name,
        project_name=metadata.project_name,
        project_path=metadata.project_path,
        project_version=metadata.project_version,
        source_dir=Path(metadata.project_path).expanduser(),
    )

    _add_src_to_path(metadata.source_dir, project_path)
    configure_project(metadata.package_name)
    return metadata



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Any, Dict, Generator, List, MutableMapping, Optional

import mlflow
from kedro.extras.datasets.json import JSONDataSet
from kedro.framework.hooks import hook_impl
from kedro.io import DataCatalog, DataSetError
from kedro.pipeline.node import Node

logger = logging.getLogger("eensight")


def _flatten_dict(params: Dict[Any, Any], delimiter: str = "/") -> Dict[str, Any]:
    """Flatten hierarchical dict, e.g. ``{'a': {'b': 'c'}} -> {'a/b': 'c'}``.

    Args:
        params (dict): Dictionary containing the hyperparameters.
        delimiter (str): Delimiter to express the hierarchy. Defaults to ``'/'``.

    Returns:
        Flattened dict.

    Examples:
        >>> _flatten_dict({'a': {'b': 'c'}})
        {'a/b': 'c'}
        >>> _flatten_dict({'a': {'b': 123}})
        {'a/b': 123}
        >>> _flatten_dict({5: {'a': 123}})
        {'5/a': 123}
    """

    def _dict_generator(
        input_dict: Any, prefixes: List[Optional[str]] = None
    ) -> Generator[Any, Optional[List[str]], List[Any]]:
        prefixes = prefixes[:] if prefixes else []
        if isinstance(input_dict, MutableMapping):
            for key, value in input_dict.items():
                key = str(key)
                if isinstance(value, MutableMapping):
                    yield from _dict_generator(value, prefixes + [key])
                else:
                    yield prefixes + [key, value if value is not None else str(None)]
        else:
            yield prefixes + [input_dict if input_dict is None else str(input_dict)]

    return {delimiter.join(keys): val for *keys, val in _dict_generator(params)}


class ActivityHooks:
    @property
    def _logger(self):
        return logging.getLogger("eensight")

    @hook_impl
    def before_node_run(
        self,
        node: Node,
        catalog: DataCatalog,
        inputs: Dict[str, Any],
        is_async: bool,
        session_id: str,
    ):

        if catalog.load("app").get("namespace") == "apply":
            if "apply.activity" in inputs:
                try:
                    activity = catalog.load("apply.activity-adjusted")
                except DataSetError:
                    return
                else:
                    self._logger.warning(
                        "Adjusted activity was found in "
                        "catalog and selected as activity feature"
                    )
                    return {**inputs, **{"apply.activity": activity}}

            if node.name == "apply.create_activity_feature":
                try:
                    activity = catalog.load("apply.activity-adjusted")
                except DataSetError:
                    self._logger.warning("bad")
                    return
                else:
                    return {
                        **inputs,
                        **{"params:activity.adjusted_activity": activity},
                    }


class MLflowHooks:
    @hook_impl
    def before_node_run(
        self,
        node: Node,
        catalog: DataCatalog,
        inputs: Dict[str, Any],
        is_async: bool,
        session_id: str,
    ) -> None:
        """Hook to be invoked before a node runs.
        This hook logs in mlflow all the input parameters of the nodes.

        Args:
            node (Node): The ``Node`` to run.
            catalog (DataCatalog): A ``DataCatalog`` containing the node's inputs
                and outputs.
            inputs (dict): The dictionary of node's input dataset. The keys are
                dataset names and the values are the actual loaded input data, not
                the dataset instance.
            is_async (bool): Whether the node was run in ``async`` mode.
            session_id: The id of the session.
        """

        run_id = catalog.load("app").get("run_id")
        if run_id is not None:
            namespace = catalog.load("app").get("namespace")

            with mlflow.start_run(run_id=run_id, nested=True):
                # log parameters
                params = {}
                for k, v in inputs.items():
                    # detect parameters automatically based on kedro reserved names
                    if k.startswith("params:"):
                        params[k.split(":")[-1]] = v

                params = _flatten_dict(params)
                for k, v in params.items():
                    mlflow.log_param("_".join([namespace, k]), v)

    @hook_impl
    def after_node_run(
        self,
        node: Node,
        catalog: DataCatalog,
        inputs: Dict[str, Any],
        outputs: Dict[str, Any],
        is_async: bool,
        session_id: str,
    ) -> None:
        """Hook to be invoked after a node runs.
        This hook logs in mlflow all generated metrcis.

        Args:
            node (Node): The ``Node`` that ran.
            catalog (DataCatalog): A ``DataCatalog`` containing the node's inputs
                and outputs.
            inputs (dict): The dictionary of inputs dataset. The keys are dataset
                names and the values are the actual loaded input data, not the dataset
                instance.
            outputs (dict): The dictionary of outputs dataset. The keys are dataset
                names and the values are the actual computed output data, not the dataset
                instance.
            is_async: Whether the node was run in ``async`` mode.
            session_id: The id of the session.
        """

        run_id = catalog.load("app").get("run_id")
        if run_id is not None:
            namespace = catalog.load("app").get("namespace")

            with mlflow.start_run(run_id=run_id, nested=True):
                # Log metrics
                for name, content in outputs.items():
                    dataset = catalog._get_dataset(name)
                    if isinstance(dataset, JSONDataSet):
                        for k, v in content.items():
                            mlflow.log_metric("_".join([namespace, k]), v)



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Literal, Union

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OrdinalEncoder
from sklearn.utils.validation import check_is_fitted

from eensight.utils import as_list, get_categorical_cols

UNKNOWN_VALUE = -1


class SafeOrdinalEncoder(TransformerMixin, BaseEstimator):
    """Encode categorical features as an integer array.

    The encoder converts the categorical features into ordinal integers. This results
    in a single column of integers (0 to n_categories - 1) per feature.

    Args:
        features (str or list of str, optional): The names of the columns to encode. If
            None, all categorical columns will be encoded. Defaults to None.
        int_is_categorical (bool, optional): If True, integer types are considered categorical.
            Defaults to False.
        unknown_value (int, optional): This parameter will set the encoded value for unknown
            categories. It has to be distinct from the values used to encode any of the categories
            in `fit`. If None, the value `-1` is used. During `transform`, unknown categories will
            be replaced using the most frequent value along each column. Defaults to None.
        remainder ({'drop', 'passthrough'}, optional): By specifying ``remainder='passthrough'``,
            all the remaining columns of the input dataset will be automatically passed through
            (concatenated with the output of the transformer), otherwise, they will be dropped.
            Defaults to "passthrough".
    """

    def __init__(
        self,
        *,
        features: Union[str, List[str]] = None,
        int_is_categorical: bool = False,
        unknown_value: int = None,
        remainder: Literal["drop", "passthrough"] = "passthrough",
    ):
        if remainder not in ("passthrough", "drop"):
            raise ValueError('Parameter "remainder" should be "passthrough" or "drop"')

        self.features = features
        self.int_is_categorical = int_is_categorical
        self.unknown_value = unknown_value
        self.remainder = remainder

    def fit(self, X: pd.DataFrame, y=None):
        """Fit the encoder on the available data.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The features dataframe.
            y (None, optional): Ignored. Defaults to None.

        Returns:
            SafeOrdinalEncoder: Fitted encoder.
        """
        if self.features is not None:
            cat_features = as_list(self.features)
            for name in cat_features:
                if pd.api.types.is_float_dtype(X[name]):
                    raise ValueError(
                        f"The encoder is applied on numerical data `{name}`"
                    )
        else:
            cat_features = get_categorical_cols(
                X, int_is_categorical=self.int_is_categorical
            )

        if X[cat_features].isnull().values.any():
            raise ValueError("Found NaN values in input's categorical data")

        self.feature_pipeline_ = Pipeline(
            [
                (
                    "select",
                    ColumnTransformer(
                        [("select", "passthrough", cat_features)], remainder="drop"
                    ),
                ),
                (
                    "encode_ordinal",
                    OrdinalEncoder(
                        handle_unknown="use_encoded_value",
                        unknown_value=self.unknown_value or UNKNOWN_VALUE,
                        dtype=np.int16,
                    ),
                ),
                (
                    "impute_unknown",
                    SimpleImputer(
                        missing_values=self.unknown_value or UNKNOWN_VALUE,
                        strategy="most_frequent",
                    ),
                ),
            ]
        )
        # Fit the pipeline
        self.feature_pipeline_.fit(X)
        self.cat_features_ = cat_features
        self.fitted_ = True
        return self

    def transform(self, X: pd.DataFrame):
        """Apply the encoder.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The features
                dataframe.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.

        Returns:
            pandas.DataFrame: The encoded features dataframe.
        """
        check_is_fitted(self, "fitted_")
        out = pd.DataFrame(
            data=self.feature_pipeline_.transform(X),
            columns=self.cat_features_,
            index=X.index,
        )

        if self.remainder == "passthrough":
            columns = X.columns
            X = X.drop(self.cat_features_, axis=1)
            out = pd.concat((X, out), axis=1).reindex(columns=columns)

        return out



LOCAL_FILE_URI_PREFIX = "file:"

SITE_ID_HELP = """The id or name of the building site. This enables path configurations
that include the site id to differentiate between different datasets."""

STORE_URI_HELP = """The URI where the generated data and models should be stored. If not
provided, the function will search for an `EENSIGHT_STORE_URI` environmemt variable."""

NAMESPACE_HELP = """There are three (3) namespaces in eensight: train, test, apply.
Datasets and pipelines that exist in the train namespace will be used for baseline model
training. Datasets and pipelines that exist in the test namespace will be used for baseline
model evaluation. Datasets and pipelines that exist in the apply namespace will be used for
counterfactual prediction (during reporting period). The default value is `train`."""

AUTO_HELP = """Flag to indicate that out-of-sample and counterfactual predictions should
use the autoencoding approach."""

INPUT_URI_HELP = """The URI where the input data can be found. If provided, the input features
are expected to be found at `{{ input_uri }}/{{ namespace }}/features` and the input labels
at `{{ input_uri }}/{{ namespace }}/labels`. If not provided, eensight will use
`{{ store_uri }}/{{ site_id }}/01_raw/{{ namespace }}/features` path as default for the input
features and `{{ store_uri }}/{{ site_id }}/01_raw/{{ namespace }}/labels` for the input labels."""

BATCH_HELP = """The batch id for the input data (features and labels) to apply the pipelines on.
In eensight, training and testing of pipelines (`train` and `test` namespaces) is done on specific
datasets, but applying the pipelines (`apply` namespace) can be done on different batches of the data.
If `batch` is provided, the input data is expected to be inside a `.../apply/{{ batch }}/features`
path for features and `.../apply/{{ batch }}/labels` for labels. If `--namespace` is not `apply`,
this option will be ignored since no `apply` pipeline (i.e. pipeline that operates on `apply` data)
will run. The default value is `None`."""

DISABLE_TRACKING_HELP = """Flag to disable MLflow tracking. By default, MLflow tracking is
enabled, which means: (a) all runs get a unique id (run_id), (b) all artifacts generated by
a run are saved in a `{{ store_uri }}/{{ site_id }}/tracked/{{ run_id}}/` path, and (c) all
parameters and metrics are stored by the MLflow tracking server. If disabled, (a) runs do not
get a run_id, (b) all artifacts generated are saved in a `{{ store_uri }}/{{ site_id }}/untracked/`
path, and (c) parameters and metrics are not stored by MLflow. If namespace is `apply`, MLflow
tracking is disabled automatically."""

EXP_NAME_HELP = """The name of the experiment for experiment tracking by MLflow. If not provided,
the MLflow default will be used. This argument will be ignored if MLflow tracking is disabled."""

TRACK_URI_HELP = """The URI of the local or remote MLflow tracking server to persist experiment
metrics. Acceptable URIs are SQLAlchemy-compatible database connection strings (e.g.
'sqlite:///path/to/file.db') or local filesystem URIs. If not provided, defaults to the
`MLFLOW_TRACKING_URI` environment variable if set, otherwise to `{{ store_uri }}/{{ site_id }}/mlruns`
if `store_uri` is a local directory."""

RUN_ID_HELP = """The ID string for the run that contains the artifacts to be used (by `test` and
`apply` pipelines) or updated (by `train` pipelines). If not provided and namespace is `train` and
MLflow tracking is enabled, a new run ID will be created. If not provided and namespace is `test`
or `apply`, the untracked artifacts will be used."""

FORMAT_HELP = (
    lambda x: f"""The file format of the dataset that contains the input {x}. The supported
file formats are csv, parquet and json. The default value is `csv`."""
)

FEATURES_FORMAT_HELP = FORMAT_HELP("features")
LABELS_FORMAT_HELP = FORMAT_HELP("labels")

LOAD_HELP = (
    lambda x, y: f"""The arguments for loading the dataset that contains the input {x}. For csv files,
the arguments are passed to pandas.read_csv (https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html),
for json files to pandas.read_json (https://pandas.pydata.org/docs/reference/api/pandas.read_json.html) and
for parquet to pandas.read_parquet (https://pandas.pydata.org/docs/reference/api/pandas.read_parquet.html).
This option can be used multiple times, and each time the expected format is `--{y} key=value`. As an
example, `--{y} sep=,`."""
)

FEATURES_LOAD_HELP = LOAD_HELP("features", "f-load-arg")
LABELS_LOAD_HELP = LOAD_HELP("labesl", "l-load-arg")

PARAMS_ARG_HELP = """One or more extra parameters for the run. This option can be used
multiple times. Each parameter should be provided in `key=value` format. Hierarchical
structure can be defined by using dot-lists, such as '--param a.bb.ccc=1'."""

FROM_INPUTS_HELP = """A comma-separated list of artifact names which should be used as
the starting point of the run."""

TO_OUTPUTS_HELP = """A comma-separated list of artifact names which should be used as
the end point."""

FROM_NODES_HELP = """A comma-separated list of node names which should be used as the
starting point."""

TO_NODES_HELP = """A comma-separated list of node names which should be used as the
end point."""

NODE_ARG_HELP = """Run only nodes with specified names. This option can be used multiple
times."""

RUNNER_ARG_HELP = """Specify a runner that you want to run the pipeline with. Available
runners: `SequentialRunner`, `ParallelRunner` and `ThreadRunner`. The default is
`SequentialRunner`."""

TAG_ARG_HELP = """One or more tags for the run. The tags are logged in the MLflow tracking
server, so they will be ignored if MLflow tracking is disabled. They should be provided in
`key=value` format."""

CONFIG_FILE_HELP = """A YAML configuration file to load the run command arguments from. The
values provided by this file will fill in any missing arguments of the run command with the
expection of the `run_id` argument. This should be provided only through the command line. If
command line arguments are also provided, they will override the ones inside the configuration file."""



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import logging

import pandas as pd

from eensight.methods.prediction.metrics import cvrmse, nmbe

logger = logging.getLogger("eensight")


def evaluate_on_test(labels: pd.DataFrame, prediction: pd.DataFrame):
    """Evaluate predictions on `test` data.

    Args:
        labels (pd.DataFrame): The target data.
        prediction (pd.DataFrame): The prediction data.

    Returns:
        dict: The out-of-sample CV(RMSE) and NMBE metrics.
    """

    metrics = {
        "CVRMSE": cvrmse(labels, prediction),
        "NMBE": nmbe(labels, prediction),
    }
    logger.info(f"Out-of-sample metrics: {metrics}")
    return metrics


def evaluate_on_test_autoenc(
    labels: pd.DataFrame,
    prediction: pd.DataFrame,
    activity: pd.DataFrame,
    assume_hurdle: bool,
):
    """Evaluate predictions on `test` data.

    Args:
        labels (pd.DataFrame): The target data.
        prediction (pd.DataFrame): The prediction data.
        activity (pd.DataFrame): The estimated activity levels.
        assume_hurdle (bool): If True, activity estimation will assume that energy
            consumption has been generated by a hurdle model.

    Returns:
        dict: The out-of-sample CV(RMSE) and NMBE metrics.
    """
    if assume_hurdle:
        selected = activity["activity"] == 1
        labels = labels[selected]
        prediction = prediction[selected]

    metrics = evaluate_on_test(labels, prediction)
    return metrics


def evaluate_on_apply(labels: pd.DataFrame, prediction: pd.DataFrame):
    """Evaluate predictions on `apply` data.

    Args:
        labels (pd.DataFrame): The target data.
        prediction (pd.DataFrame): The prediction data.

    Returns:
        pandas.Series: The estimated impact.
    """

    impact = prediction - labels
    return impact


def evaluate_on_apply_autoenc(
    labels: pd.DataFrame,
    prediction: pd.DataFrame,
    activity: pd.DataFrame,
    assume_hurdle: bool,
):
    """Evaluate predictions on `apply` data.

    Args:
        labels (pd.DataFrame): The target data.
        prediction (pd.DataFrame): The prediction data.
        activity (pd.DataFrame): The estimated activity levels.
        assume_hurdle (bool): If True, activity estimation will assume that energy
            consumption has been generated by a hurdle model.

    Returns:
        pandas.Series: The estimated impact.
    """
    impact = prediction - labels
    if assume_hurdle:
        impact = impact["consumption"].mul(activity["activity"]).to_frame("consumption")

    return impact



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path

import click
from kedro.framework.project import configure_project

from .framework.cli.run import run

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(context_settings=CONTEXT_SETTINGS, name="eensight")
def cli():
    """Command line tool for running the eetracker package."""
    pass


cli.add_command(run)


def main():
    package_name = Path(__file__).parent.name
    configure_project(package_name)
    cli()



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
import warnings
from pathlib import Path
from typing import Dict
from urllib.parse import urlparse

import click
import mlflow
from click_help_colors import HelpColorsGroup
from kedro.framework.cli.utils import env_option
from kedro.framework.session import KedroSession
from kedro.utils import load_obj
from mlflow.tracking import MlflowClient
from mlflow.tracking.context.registry import resolve_tags
from omegaconf import OmegaConf

from eensight.framework.startup import bootstrap_project
from eensight.settings import PROJECT_PATH

from .help import (
    AUTO_HELP,
    BATCH_HELP,
    CONFIG_FILE_HELP,
    DISABLE_TRACKING_HELP,
    EXP_NAME_HELP,
    FEATURES_FORMAT_HELP,
    FEATURES_LOAD_HELP,
    FROM_INPUTS_HELP,
    FROM_NODES_HELP,
    INPUT_URI_HELP,
    LABELS_FORMAT_HELP,
    LABELS_LOAD_HELP,
    LOCAL_FILE_URI_PREFIX,
    NAMESPACE_HELP,
    NODE_ARG_HELP,
    PARAMS_ARG_HELP,
    RUN_ID_HELP,
    RUNNER_ARG_HELP,
    SITE_ID_HELP,
    STORE_URI_HELP,
    TAG_ARG_HELP,
    TO_NODES_HELP,
    TO_OUTPUTS_HELP,
    TRACK_URI_HELP,
)

logger = logging.getLogger("eensight")
warnings.filterwarnings("ignore", category=DeprecationWarning)


##########################################################################################
# Utility functions
##########################################################################################


def _uri_to_string(uri):
    if uri and (urlparse(uri).scheme == "file"):
        return uri
    if uri and os.path.exists(uri):
        return Path(uri).absolute().resolve().as_posix()
    return uri


def _tuple_to_dict(ctx, param, values) -> Dict[str, str]:
    """Reformat data structure from tuple to dictionary"""
    if values:
        return dict(map(lambda x: x.split("="), values))
    return dict()


def _split_string(ctx, param, value):
    """Split string by comma."""
    return [item.strip() for item in value.split(",") if item.strip()]


def _run_kedro_session(env, extra_params, run_args):
    bootstrap_project(PROJECT_PATH)
    with KedroSession.create(
        package_name="eensight",
        project_path=PROJECT_PATH,
        env=env,
        extra_params=extra_params,
        save_on_close=False,
    ) as session:
        session.run(**run_args)


##########################################################################################
# CLI
##########################################################################################


@click.group(
    name="eensight",
    cls=HelpColorsGroup,
    help_headers_color="green",
    help_options_color="yellow",
)
def cli_run():
    pass


@cli_run.command("run", help="Run a eensight pipeline, where PIPELINE is its name.")
@click.argument("pipeline", type=str)
@click.option("--site-id", "-si", type=str, default=None, help=SITE_ID_HELP)
@click.option(
    "--store-uri",
    "-su",
    type=str,
    default=None,
    envvar="EENSIGHT_STORE_URI",
    help=STORE_URI_HELP,
)
@click.option(
    "--namespace",
    "-ns",
    type=click.Choice(["train", "test", "apply"]),
    default=None,
    help=NAMESPACE_HELP,
)
@click.option(
    "--autoencode",
    "-ac",
    is_flag=True,
    show_default=True,
    default=False,
    help=AUTO_HELP,
)
@click.option(
    "--input-uri",
    "-iu",
    type=str,
    default=None,
    help=INPUT_URI_HELP,
)
@click.option("--batch", "-b", type=str, default=None, help=BATCH_HELP)
@click.option(
    "--no-tracking",
    "-nt",
    "disable_tracking",
    is_flag=True,
    help=DISABLE_TRACKING_HELP,
)
@click.option(
    "--experiment",
    "-ex",
    "experiment_name",
    type=str,
    default=None,
    help=EXP_NAME_HELP,
)
@click.option(
    "--tracking-uri",
    "-tu",
    type=str,
    default=None,
    envvar="MLFLOW_TRACKING_URI",
    help=TRACK_URI_HELP,
)
@click.option(
    "--run-id",
    "-ri",
    type=str,
    default=None,
    help=RUN_ID_HELP,
)
@click.option(
    "--features-format",
    "-ff",
    type=click.Choice(["csv", "parquet", "json"]),
    show_default=True,
    default="csv",
    help=FEATURES_FORMAT_HELP,
)
@click.option(
    "--labels-format",
    "-lf",
    type=click.Choice(["csv", "parquet", "json"]),
    show_default=True,
    default="csv",
    help=LABELS_FORMAT_HELP,
)
@click.option(
    "--f-load-arg",
    "-fla",
    "features_load_args",
    type=str,
    multiple=True,
    callback=_tuple_to_dict,
    help=FEATURES_LOAD_HELP,
)
@click.option(
    "--l-load-arg",
    "-lla",
    "labels_load_args",
    type=str,
    multiple=True,
    callback=_tuple_to_dict,
    help=LABELS_LOAD_HELP,
)
@env_option
@click.option(
    "--param",
    "-p",
    "extra_params",
    type=click.UNPROCESSED,
    multiple=True,
    help=PARAMS_ARG_HELP,
)
@click.option(
    "--from-inputs",
    "-fi",
    type=click.UNPROCESSED,
    default="",
    help=FROM_INPUTS_HELP,
    callback=_split_string,
)
@click.option(
    "--to-outputs",
    "-to",
    type=click.UNPROCESSED,
    default="",
    help=TO_OUTPUTS_HELP,
    callback=_split_string,
)
@click.option(
    "--from-nodes",
    "-fn",
    type=click.UNPROCESSED,
    default="",
    help=FROM_NODES_HELP,
    callback=_split_string,
)
@click.option(
    "--to-nodes",
    "-tn",
    type=click.UNPROCESSED,
    default="",
    help=TO_NODES_HELP,
    callback=_split_string,
)
@click.option("--node", "-n", "node_names", type=str, multiple=True, help=NODE_ARG_HELP)
@click.option(
    "--runner",
    "-r",
    type=str,
    default=None,
    help=RUNNER_ARG_HELP,
)
@click.option(
    "--tag",
    "-t",
    "tags",
    type=str,
    multiple=True,
    callback=_tuple_to_dict,
    help=TAG_ARG_HELP,
)
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    default=None,
    help=CONFIG_FILE_HELP,
)
def run(
    pipeline: str,
    site_id: str,
    store_uri: str,
    namespace: str,
    autoencode: bool,
    input_uri: str,
    batch: str,
    disable_tracking: str,
    experiment_name: str,
    tracking_uri: str,
    run_id: str,
    features_format: str,
    labels_format: str,
    features_load_args: dict,
    labels_load_args: dict,
    env: str,
    extra_params: list,
    from_inputs: list,
    to_outputs: list,
    from_nodes: list,
    to_nodes: list,
    node_names: list,
    runner: str,
    tags: dict,
    config: Path,
):
    """Run the eensight pipelines using the provided input data."""

    if (pipeline == "evaluate") and (namespace == "train"):
        msg = "The `evaluate` pipeline supports only the `test` and `apply` namespaces."
        click.secho(click.style(msg, fg="bright_red"))
        raise click.Abort()

    if (pipeline == "adjust") and (namespace != "apply"):
        msg = "The `adjust` pipeline supports only the `apply` namespace."
        click.secho(click.style(msg, fg="bright_red"))
        raise click.Abort()

    config = OmegaConf.load(config) if config else {}

    env = env or config.get("env") or "local"
    site_id = site_id or config.get("site_id")

    if site_id is None:
        msg = (
            "`site_id` must be provided through either the cli or the `--config` file."
        )
        click.secho(click.style(msg, fg="bright_red"))
        raise click.Abort()

    store_uri = _uri_to_string(store_uri or config.get("store_uri"))
    if store_uri is None:
        msg = "`store_uri` must be provided through either the cli or the `--config` file."
        click.secho(click.style(msg, fg="bright_red"))
        raise click.Abort()

    input_uri = _uri_to_string(input_uri or config.get("input_uri"))
    namespace = namespace or config.get("namespace") or "train"

    disable_tracking = (
        disable_tracking
        or (namespace == "apply")
        or ((namespace == "test") and (run_id is None))
    )

    if not disable_tracking:
        experiment_name = experiment_name or config.get("experiment_name") or "Default"
        tracking_uri = tracking_uri or config.get("tracking_uri")

        if tracking_uri:
            tracking_uri = _uri_to_string(tracking_uri)
        elif os.path.exists(store_uri):
            tracking_uri = f"{LOCAL_FILE_URI_PREFIX}/{store_uri}/{site_id}/mlruns"
        else:
            msg = "`tracking_uri` must be provided if `store_uri` is not a local directory."
            click.secho(click.style(msg, fg="bright_red"))
            raise click.Abort()

        mlflow.set_tracking_uri(tracking_uri)
        client = MlflowClient()
        expt = client.get_experiment_by_name(experiment_name)
        if expt is not None:
            experiment_id = expt.experiment_id
        else:
            logger.info(
                f"Experiment with name {experiment_name} not found. Creating it."
            )
            experiment_id = client.create_experiment(name=experiment_name)

        tags = {
            **config.get("tags", {}),
            **tags,
            "site_id": site_id,
            "namespace": namespace,
            "autoencoding": autoencode,
        }

    features_load_args = features_load_args or config.get("features_load_args", {})
    labels_load_args = labels_load_args or config.get("labels_load_args", {})

    for load_args in [features_load_args, labels_load_args]:
        if ("sep" in load_args) and (load_args["sep"] == ","):
            del load_args["sep"]

    extra_params = [
        "=".join([p.split("=")[0].replace("-", "_"), p.split("=")[1]])
        for p in extra_params
    ]
    extra_params = {
        **config.get("extra_params", {}),
        **OmegaConf.to_container(OmegaConf.from_dotlist(extra_params)),
    }

    for key in extra_params.keys():
        if key in ("globals", "app"):
            msg = (
                f"`{key}` is a reserved word and cannot be used as a key in `--param`."
            )
            click.secho(click.style(msg, fg="bright_red"))
            raise click.Abort()

    extra_params.update(
        {
            "app": {
                "store_uri": store_uri,
                "input_uri": input_uri,
                "site_id": site_id,
                "namespace": namespace,
                "run_id": None,
                "batch": batch,
                "features": {
                    "format": features_format,
                    "load_args": features_load_args,
                },
                "labels": {"format": labels_format, "load_args": labels_load_args},
            }
        }
    )

    runner = load_obj(
        runner or config.get("runner") or "SequentialRunner", "kedro.runner"
    )
    run_args = {
        "from_nodes": from_nodes,
        "to_nodes": to_nodes,
        "from_inputs": from_inputs,
        "to_outputs": to_outputs,
        "node_names": node_names,
        "runner": runner(is_async=True),
        "pipeline_name": pipeline,
        "tags": [
            namespace,
            "_".join([namespace, "autoenc" if autoencode else "default"]),
        ],
    }

    run_args = {
        key: config.get(key) if not value else value for key, value in run_args.items()
    }

    if not disable_tracking:
        with mlflow.start_run(
            experiment_id=experiment_id, run_id=run_id, tags=resolve_tags(tags)
        ) as run:
            run_id = run.info.run_id
            run_name = run.info.run_name
            extra_params["app"]["run_id"] = run_id
            _run_kedro_session(env, extra_params, run_args)
    else:
        run_id = None
        run_name = None
        _run_kedro_session(env, extra_params, run_args)

    click.echo(f"Run ID: {run_id}, Run name (for MLflow UI): {run_name}")



"""Project pipelines."""
from typing import Dict

from kedro.pipeline import Pipeline

from .pipelines import adjust, evaluate, predict, preprocess


def register_pipelines() -> Dict[str, Pipeline]:
    """Register the project's pipelines.

    Returns:
        A mapping from a pipeline name to a ``Pipeline`` object.
    """
    preprocess_ppl = preprocess.create_pipeline()
    predict_ppl = predict.create_pipeline()
    evaluate_ppl = evaluate.create_pipeline()
    adjust_ppl = adjust.create_pipeline()

    return {
        "preprocess": preprocess_ppl,
        "predict": predict_ppl,
        "evaluate": evaluate_ppl,
        "adjust": adjust_ppl,
        "all": preprocess_ppl + predict_ppl + evaluate_ppl + adjust_ppl,
        "__default__": preprocess_ppl + predict_ppl,
    }



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
from sklearn.utils.validation import column_or_1d


def cvrmse(y_true: np.ndarray, y_pred: np.ndarray):
    """Compute the Coefficient of Variation of the Root Mean Squared Error.

    Args:
        y_true (np.ndarray): Ground truth (correct) target values.
        y_pred (np.ndarray): Estimated target values.

    Returns:
        float: The metric's value.
    """

    y_true = column_or_1d(y_true)
    y_pred = column_or_1d(y_pred)

    resid = y_true - y_pred
    return float(np.sqrt((resid**2).sum() / len(resid)) / np.mean(y_true))


def nmbe(y_true: np.ndarray, y_pred: np.ndarray):
    """Compute the Normalized Mean Bias Error.

    Args:
        y_true (np.ndarray): Ground truth (correct) target values.
        y_pred (np.ndarray): Estimated target values.

    Returns:
        float: The metric's value.
    """

    y_true = column_or_1d(y_true)
    y_pred = column_or_1d(y_pred)

    resid = y_true - y_pred
    return float(np.mean(resid) / np.mean(y_true))



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import datetime
import logging

import numpy as np
import pandas as pd

logger = logging.getLogger("eensight")


def expand_dates(data: pd.DataFrame) -> pd.DataFrame:
    """Add missing timestamps.

    If there are missing timestamps, they are added and the respective data is treated
    as missing values.

    Args:
        data (pandas.DataFrame): The data to correct. The dataframe's index must be a
            pandas DatetimeIndex and in increasing order.

    Returns:
        pandas.DataFrame: The input data with the missing timestamps added.
    """

    if not isinstance(data.index, pd.DatetimeIndex):
        raise ValueError("The input dataframe's index must be a pandas DatetimeIndex.")

    if not data.index.is_monotonic_increasing:
        raise ValueError("The input dataframe's index must be in increasing order.")

    dt = data.index.to_series().diff()
    time_step = dt.iloc[dt.values.nonzero()[0]].min()

    if time_step >= pd.Timedelta(days=1):
        return data

    if time_step == pd.Timedelta("0 days 00:00:00"):
        raise ValueError("Input data contains dublicate dates")

    full_index = pd.date_range(
        start=datetime.datetime.combine(data.index.min().date(), datetime.time(0, 0)),
        end=datetime.datetime.combine(
            data.index.max().date() + datetime.timedelta(days=1), datetime.time(0, 0)
        ),
        freq=time_step,
        inclusive="left",
    )
    return data.reindex(full_index)


def filter_data(
    X: pd.Series,
    *,
    min_value: float = None,
    max_value: float = None,
    allow_zero: bool = True,
    allow_negative: bool = False,
    copy=True,
) -> pd.Series:
    """Screen for non-physically plausible values in the data.

    Args:
        X (pandas Series): Time series data to filter.
        min_value (float, optional): Minumun acceptable value. Values lower than
            `min_value` will be replaced by NaN. Defaults to None.
        max_value (float, optional): Maximum acceptable value. Values greater than
            `max_value` will be replaced by NaN. Defaults to None.
        allow_zero (bool, optional): Whether zero values should be allowed. Defaults
            to True.
        allow_negative (bool, optional): Whether negative values should be allowed.
            Defaults to False.
        copy (bool, optional): Whether the NaN replacements should performed on a copy
            of the input data. Defaults to True.

    Raises:
        ValueError: If input data is not pandas Series.

    Returns:
        pandas.Series: The filtered data.
    """
    if not isinstance(X, pd.Series):
        raise ValueError("Input data is expected of pandas Series")

    if copy:
        X = X.copy()

    nan_before = X.isna().sum()

    if min_value is not None:
        X.loc[X < min_value] = np.nan
    if max_value is not None:
        X.loc[X > max_value] = np.nan
    if not allow_zero:
        X.loc[X.abs() <= 1e-05] = np.nan
    if not allow_negative:
        X.loc[X < 0] = np.nan

    nan_after = X.isna().sum()
    if nan_after > nan_before:
        logger.info(f"{nan_after - nan_before} NaN values added by filtering.")

    return X


def check_data_adequacy(
    features: pd.DataFrame, labels: pd.DataFrame, max_missing_pct: float = 0.1
) -> pd.DataFrame:
    """Check whether enough data is available for each month of the year.

    Args:
        features (pandas.DataFrame of shape (n_samples, n_features)): The features dataframe.
        labels (pandas.DataFrame of shape (n_samples, 1)): The labels dataframe.
        max_missing_pct (float, optional): The maximum acceptable percentage of missing
            data per month. Defaults to 0.1.

    Raises:
        ValueError: If `max_missing_pct` is larger than 0.99 or less than 0.

    Returns:
        pandas.DataFrame: Dataframe containing percentage of missing data per month.
    """
    if (max_missing_pct > 0.99) or (max_missing_pct < 0):
        raise ValueError(
            "Values for `max_missing_pct` cannot be larger than "
            "0.99 or smaller than 0"
        )

    missing = labels.mask(features.isna().all(axis=1), np.nan)
    missing_per_month = dict()

    for month, group in missing.groupby(lambda x: x.month):
        missing_per_month[month] = (
            np.sum(
                group.groupby([lambda x: x.day, lambda x: x.hour]).count() == 0
            ).item()
            / 720  # hours per month
        )

    missing_per_month = {f"M{key}": val for key, val in missing_per_month.items()}
    missing_per_month = pd.DataFrame.from_dict(
        missing_per_month, orient="index", columns=["missing_pct"]
    )

    insufficient = missing_per_month[missing_per_month["missing_pct"] > max_missing_pct]

    if len(insufficient) > 0:
        logger.warning(
            f"Months with not enough data are:\n {insufficient['missing_pct'].to_dict()}"
        )
    return missing_per_month



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import warnings
from functools import partial
from typing import List, Union

import numpy as np
import optuna
import pandas as pd
import statsmodels.api as sm
from catboost import CatBoostRegressor
from sklearn.base import BaseEstimator
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error
from sklearn.preprocessing import KBinsDiscretizer, MinMaxScaler, SplineTransformer
from sklearn.utils.validation import check_is_fitted, column_or_1d
from statsmodels.tools.sm_exceptions import ConvergenceWarning
from tsmoothie.smoother import KalmanSmoother

from eensight.methods.prediction.metrics import cvrmse
from eensight.methods.preprocessing.alignment import get_time_step
from eensight.utils import as_list, get_categorical_cols, tensor_product

warnings.filterwarnings("ignore", category=optuna.exceptions.ExperimentalWarning)
optuna.logging.set_verbosity(optuna.logging.ERROR)


BOOST_PARAMS = {
    "learning_rate": 0.01,
    "depth": 3,
    "bootstrap_type": "Bayesian",
    "early_stopping_rounds": 100,
    "task_type": "CPU",
    "has_time": False,
    "allow_writing_files": False,
    "verbose": False,
}


##################################################################################
# Utilities
##################################################################################


def _sigmoid(X, c1, c2):
    return 1 / (1 + np.exp(-c1 * (X - c2)))


def _rescale(X, scale_range=(0, 1)):
    scaler = MinMaxScaler(feature_range=scale_range)

    return pd.Series(
        scaler.fit_transform(column_or_1d(X).reshape(-1, 1)).squeeze(), index=X.index
    )


def _select_c2(trial, X, y, activity):
    c1 = trial.suggest_float("c1", 5, 100, step=1)
    c2 = trial.suggest_float("c2", 0.1, 0.9)

    activity_trimmed = (
        activity.pipe(_sigmoid, c1, c2)
        .pipe(_rescale)
        .map(lambda x: 0 if x < 0.1 else 1 if x > 0.9 else x)
    )

    dmatrix = SplineTransformer(n_knots=4, degree=3, knots="quantile").fit_transform(X)

    prediction = []
    for weights in [activity_trimmed, 1 - activity_trimmed]:
        prediction.append(
            pd.Series(
                LinearRegression(fit_intercept=False)
                .fit(dmatrix, y, sample_weight=weights)
                .predict(dmatrix)
                .squeeze(),
                index=y.index,
            )
            * weights
        )

    prediction = np.sum(prediction, axis=0)
    return mean_squared_error(y["consumption"], prediction)


def _overfit_proxy(X, y, y_eval, activity, c1, c2):
    activity_trimmed = (
        activity.pipe(_sigmoid, c1, c2)
        .pipe(_rescale)
        .map(lambda x: 0 if x < 0.1 else 1 if x > 0.9 else round(x, 1))
    )

    dmatrix = tensor_product(
        SplineTransformer(n_knots=3, degree=3, knots="quantile").fit_transform(
            activity_trimmed.values.reshape(-1, 1)
        ),
        SplineTransformer(n_knots=4, degree=3, knots="quantile").fit_transform(
            X.values
        ),
    )

    prediction = pd.Series(
        LinearRegression(fit_intercept=False)
        .fit(dmatrix, y)
        .predict(dmatrix)
        .squeeze(),
        index=y.index,
    )
    return np.abs(cvrmse(y_eval, prediction) - cvrmse(y, prediction))


def _select_c1(trial, X, y, y_eval, activity, c2, init_value):
    c1 = trial.suggest_float("c1", 5, 100, step=1)
    proxy = _overfit_proxy(X, y, y_eval, activity, c1, c2)
    return ((init_value - proxy) / init_value) - 0.005 * c1


def _objective_adjust(trial, X, y, activity, predictor, upper_bound):
    true_bound = trial.suggest_float("true_bound", 0, upper_bound)
    c1 = trial.suggest_float("c1", 5, 50, step=1)
    c2 = trial.suggest_float("c2", 0.1, 0.9)

    activity_trimmed = (
        activity.pipe(_sigmoid, c1, c2)
        .pipe(_rescale, (0, true_bound))
        .map(lambda x: 0 if x < 0.1 else 1 if x > 0.9 else round(x, 1))
    )

    X_act = pd.concat([X, activity_trimmed.to_frame("activity")], axis=1)
    y_pred = predictor.predict(X_act)
    return mean_squared_error(y["consumption"], y_pred)


##################################################################################
# Activity estimation
##################################################################################


def estimate_activity_markov(
    X: pd.DataFrame,
    y: pd.DataFrame,
    exog: str = "temperature",
    n_bins: int = 5,
):
    """Estimate activity using a Markov switching regression model with 2 regimes.

    Args:
        X (pandas.DataFrame): The feature data.
        y (pandas.DataFrame): The label data.
        exog (str, optional): The name of the feature to use for the underlying
            univariate Markov switching dynamic regression. It should not contain
            categorical data. Defaults to "temperature".
        n_bins (int, optional): The number of bins to discretize the `feature` values.
            Defaults to 5.

    Returns:
        pandas.Series: The estimated activity time series.
    """

    if not pd.api.types.is_float_dtype(X[exog]):
        raise ValueError("`exog` feature cannot contain categorical data.")

    index = X.index
    X = X.dropna()
    y = y.loc[X.index]

    subsampled = False
    if get_time_step(index) < pd.Timedelta(hours=1):
        subsampled = True
        X = X.groupby([lambda x: x.date, lambda x: x.hour]).sample(n=1).sort_index()
        y = y.loc[X.index]

    enc = KBinsDiscretizer(n_bins=n_bins, encode="onehot-dense", strategy="uniform")
    X = enc.fit_transform(X[[exog]])

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        warnings.filterwarnings("ignore", category=ConvergenceWarning)
        results = sm.tsa.MarkovRegression(
            y.values.squeeze(), k_regimes=2, exog=X, switching_variance=False
        ).fit()

    act = pd.Series(
        results.smoothed_marginal_probabilities[:, 1], index=y.index
    ).reindex(index)

    if subsampled:
        act = (
            act.groupby([lambda x: x.date, lambda x: x.hour])
            .fillna(method="ffill")
            .fillna(method="bfill")
        )

    # round up and down to remove small variations that may be picked
    # up by a predicitive model
    return act.map(lambda x: 0 if x < 0.1 else 1 if x > 0.9 else x)


def extract_activity(
    X: pd.DataFrame,
    y: pd.DataFrame,
    non_occ_features: Union[str, List[str]] = None,
    cat_features: Union[str, List[str]] = None,
    scale_range: tuple = (0, 1),
) -> pd.Series:
    """Perform the activity estimation task.

    The function uses two (2) quantile regressions of the label data as function
    of the `features` in input `X`: one at the 0.99 quantile and one at the 0.01
    one. Then, all observations are normalized to the [0, 1] interval using the
    range of the corresponding 0.99-quantile and 0.01-quantile predictions.

    Args:
        X (pandas.DataFrame): The feature data.
        y (pandas.DataFrame): The label data.
        non_occ_features (str or list of str, optional): The names of the features
            to use to fit the two quantile regression models on the label data. If
            not provided, "temperature" will be used. Defaults to None.
        cat_features (str or list of str, optional): The name(s) of the categorical
            features in the input data. Defaults to None.
        scale_range (tuple (min, max), optional): Desired range of activity levels.
            Defaults to `(0, 1)`.

    Returns:
        pandas.Series: The estimated activity time series.
    """

    a, b = scale_range
    non_occ_features = as_list(non_occ_features) or ["temperature"]
    cat_features = as_list(cat_features)

    if cat_features:
        cat_features = [name for name in cat_features if name in non_occ_features]
    else:
        cat_features = [
            name
            for name in get_categorical_cols(X, int_is_categorical=False)
            if name in non_occ_features
        ]

    model_upper = CatBoostRegressor(
        **dict(
            BOOST_PARAMS,
            loss_function="Quantile:alpha=0.99",
            iterations=600,
            cat_features=cat_features,
        )
    )
    model_lower = CatBoostRegressor(
        **dict(
            BOOST_PARAMS,
            loss_function="Quantile:alpha=0.01",
            iterations=200,
            cat_features=cat_features,
        )
    )

    model_upper = model_upper.fit(X[non_occ_features], y["consumption"])
    model_lower = model_lower.fit(X[non_occ_features], y["consumption"])

    pred_upper = pd.Series(model_upper.predict(X[non_occ_features]), index=X.index)
    pred_lower = pd.Series(model_lower.predict(X[non_occ_features]), index=X.index)

    act = (b - a) * ((y["consumption"] - pred_lower) / (pred_upper - pred_lower)) + a
    act = act.clip(lower=a, upper=b)
    return act


def estimate_activity(
    X: pd.DataFrame,
    y: pd.DataFrame,
    non_occ_features: Union[str, List[str]] = None,
    cat_features: Union[str, List[str]] = None,
    exog: str = "temperature",
    n_trials: int = 200,
    assume_hurdle: bool = False,
    verbose: bool = False,
    return_params: bool = False,
):
    """Perform the activity estimation task.

    Args:
        X (pandas.DataFrame): The feature data.
        y (pandas.DataFrame): The label data.
        non_occ_features (str or list of str, optional): The names of the features to use
            to fit the two quantile regression models on the label data. If not provided,
            "temperature" will be used. Defaults to None.
        cat_features (str or list of str, optional): The name(s) of the categorical
            features in the input data. Defaults to None.
        exog (str, optional): The name of the feature to use for determining how to best split
            the data into two regimes. It should not contain categorical data. Defaults to
            "temperature".
        n_trials (int, optional): The number of iterations for the underlying optimization.
            Defaults to 200.
        assume_hurdle (bool, optional): If True, it is assumed that the energy consumption
            data is generated from a hurdle model. A hurdle model is a two-part model that
            specifies one process for zero values and another process for the positive ones.
            Defaults to False.
        verbose (bool, optional): If True, a progress bar is visible. Defaults to False.
        return_params (bool, optional): If True, a dictionary with the optimal parameters of
            the sigmoid transformation will be returned as second argument. Defaults to False.

    Returns:
        pandas.Series: The estimated activity time series.
    """

    cat_features = as_list(cat_features)
    cat_features = [name for name in cat_features if name in X.columns]

    if assume_hurdle:
        y_on = y[y["consumption"] > 1e-05]
        X_on = X.loc[y_on.index]

        act = (
            extract_activity(
                X_on,
                y_on,
                non_occ_features=non_occ_features,
                cat_features=cat_features,
                scale_range=(0.05, 1),
            )
            .map(lambda x: round(x, 1))
            .reindex(y.index)
            .fillna(value=0)
        )
    else:
        act = extract_activity(
            X, y, non_occ_features=non_occ_features, cat_features=cat_features
        ).map(lambda x: round(x, 1))

    X = X.dropna()
    _objective = partial(
        _select_c2,
        X=X[[exog]],
        y=y.loc[X.index],
        activity=act.loc[X.index],
    )

    study = optuna.create_study(
        sampler=optuna.samplers.TPESampler(multivariate=True),
        direction="minimize",
    )
    study.optimize(_objective, n_trials=n_trials, show_progress_bar=verbose)
    c2 = study.best_params["c2"]

    # denoise the energy consumption data
    smoother = KalmanSmoother(component="level", component_noise={"level": 0.1})
    smoother = smoother.smooth(y)
    y_sm = pd.DataFrame(smoother.smooth_data[0], index=y.index, columns=["consumption"])
    y_sm.loc[y[y["consumption"] <= 1e-05].index] = 0

    init_value = _overfit_proxy(
        X[[exog]],
        y.loc[X.index],
        y_sm.loc[X.index],
        act.loc[X.index],
        5,
        c2,
    )
    _objective = partial(
        _select_c1,
        X=X[[exog]],
        y=y.loc[X.index],
        y_eval=y_sm.loc[X.index],
        activity=act.loc[X.index],
        c2=c2,
        init_value=init_value,
    )

    study = optuna.create_study(
        sampler=optuna.samplers.TPESampler(),
        direction="maximize",
    )
    study.optimize(_objective, n_trials=n_trials, show_progress_bar=verbose)
    c1 = study.best_params["c1"]

    act = (
        act.pipe(_sigmoid, c1, c2)
        .pipe(_rescale)
        .map(lambda x: 0 if x < 0.1 else 1 if x > 0.9 else round(x, 1))
    )

    if return_params:
        return act, {"c1": c1, "c2": c2}
    return act


def adjust_activity(
    X: pd.DataFrame,
    y: pd.DataFrame,
    predictor: BaseEstimator,
    non_occ_features: Union[str, List[str]] = None,
    cat_features: Union[str, List[str]] = None,
    assume_hurdle: bool = False,
    n_trials: int = 500,
    upper_bound: float = 2,
    verbose: bool = False,
    return_params: bool = False,
):
    """Adjust activity levels for events that affect density of energy consumption.

    Args:
        X (pandas.DataFrame): The feature data.
        y (pandas.DataFrame): The label data.
        predictor (BaseEstimator): A trained model that predicts energy consumption given
            `non_occ_features` and activity levels.
        non_occ_features (str or list of str, optional): The names of the occupancy-
            independent features to use for estimating activity levels. If not provided,
            "temperature" will be used. Defaults to None.
        cat_features (str or list of str, optional): The name(s) of the categorical
            features in the input data. Defaults to None.
        assume_hurdle (bool, optional): If True, it is assumed that the energy consumption
            data is generated from a hurdle model. Defaults to False.
        n_trials (int, optional): The number of iterations for the underlying optimization.
            Defaults to 500.
        upper_bound (float, optional): The maximum value that the true upper bound for the
            adjusted activity levels can take. Defaults to 2.
        verbose (bool, optional): If True, a progress bar is visible. Defaults to False.
        return_params (bool, optional): If True, a dictionary with the optimal parameters of
            the sigmoid transformation will be returned as second argument. Defaults to False.

    Returns:
        pandas.Series: The adjusted activity time series.
    """

    check_is_fitted(predictor)
    non_occ_features = as_list(non_occ_features) or ["temperature"]

    if assume_hurdle:
        y_on = y[y["consumption"] > 1e-05]
        X_on = X.loc[y_on.index]

        activity = (
            extract_activity(
                X_on,
                y_on,
                non_occ_features=non_occ_features,
                cat_features=cat_features,
                scale_range=(0.05, 1),
            )
            .map(lambda x: round(x, 1))
            .reindex(y.index)
            .fillna(value=0)
        )
    else:
        activity = extract_activity(
            X, y, non_occ_features=non_occ_features, cat_features=cat_features
        ).map(lambda x: round(x, 1))

    _objective = partial(
        _objective_adjust,
        X=X,
        y=y,
        activity=activity,
        predictor=predictor,
        upper_bound=upper_bound,
    )
    study = optuna.create_study(
        sampler=optuna.samplers.TPESampler(multivariate=True),
    )
    study.optimize(_objective, n_trials=n_trials, show_progress_bar=verbose)

    c1 = study.best_params["c1"]
    c2 = study.best_params["c2"]
    true_bound = study.best_params["true_bound"]

    activity = (
        activity.pipe(_sigmoid, c1, c2)
        .pipe(_rescale, (0, true_bound))
        .map(lambda x: 0 if x < 0.1 else 1 if x > 0.9 else round(x, 1))
    )

    if return_params:
        return activity, {
            "c1": c1,
            "c2": c2,
            "true_bound": true_bound,
        }
    return activity



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import datetime
import decimal
import json
import logging
import sys
import traceback
from typing import Dict

import numpy as np
import pandas as pd
from pandas.api.types import is_datetime64_any_dtype as is_datetime

logger = logging.getLogger("eensight")


############################################################################################
# Utility functions

# Adapted from https://github.com/great-expectations/great_expectations
############################################################################################


def convert_to_json_serializable(data):
    """Helper function to convert an object to one that is json serializable.

    Args:
        data (Any): An object to attempt to convert a corresponding json-serializable
            object

    Returns:
        dict: The converted object.
    """
    try:
        if not isinstance(data, list) and pd.isna(data):
            return None
    except TypeError:
        pass
    except ValueError:
        pass

    if isinstance(data, (str, int, float, bool)):
        # No problem to encode json
        return data

    elif isinstance(data, dict):
        new_dict = {}
        for key in data:
            new_dict[str(key)] = convert_to_json_serializable(data[key])

        return new_dict

    elif isinstance(data, (list, tuple, set)):
        new_list = []
        for val in data:
            new_list.append(convert_to_json_serializable(val))

        return new_list

    elif isinstance(data, (np.ndarray, pd.Index)):
        return [convert_to_json_serializable(x) for x in data.tolist()]

    elif data is None:
        # No problem to encode json
        return data

    elif isinstance(data, (datetime.datetime, datetime.date)):
        return data.isoformat()

    elif np.issubdtype(type(data), np.bool_):
        return bool(data)

    elif np.issubdtype(type(data), np.integer) or np.issubdtype(type(data), np.uint):
        return int(data)

    elif np.issubdtype(type(data), np.floating):
        return float(round(data, sys.float_info.dig))

    elif isinstance(data, pd.Series):
        index_name = data.index.name or "index"
        value_name = data.name or "value"
        return [
            {
                index_name: convert_to_json_serializable(idx),
                value_name: convert_to_json_serializable(val),
            }
            for idx, val in data.iteritems()
        ]

    elif isinstance(data, pd.DataFrame):
        return convert_to_json_serializable(data.to_dict(orient="records"))

    elif isinstance(data, decimal.Decimal):
        if not (-1e-55 < decimal.Decimal.from_float(float(data)) - data < 1e-55):
            logger.warning(
                f"Using lossy conversion for decimal {data} "
                "to float object to support serialization."
            )
        return float(data)

    else:
        raise TypeError(
            f"{data} is of type {type(data).__name__} which cannot be serialized."
        )


def ensure_json_serializable(data):
    """Helper function to check if an object is json serializable.

    Args:
        data (Any): An object to attempt to convert into a json-serializable
            object.
    """
    try:
        if not isinstance(data, list) and pd.isna(data):
            return
    except TypeError:
        pass
    except ValueError:
        pass

    if isinstance(data, ((str,), (int,), float, bool)):
        # No problem to encode json
        return

    elif isinstance(data, dict):
        for key in data:
            str(key)  # key must be cast-able to string
            ensure_json_serializable(data[key])
        return

    elif isinstance(data, (list, tuple, set)):
        for val in data:
            ensure_json_serializable(val)
        return

    elif isinstance(data, (np.ndarray, pd.Index)):
        _ = [ensure_json_serializable(x) for x in data.tolist()]
        return

    elif data is None:
        # No problem to encode json
        return

    elif isinstance(data, (datetime.datetime, datetime.date)):
        return

    elif np.issubdtype(type(data), np.bool_):
        return

    elif np.issubdtype(type(data), np.integer) or np.issubdtype(type(data), np.uint):
        return

    elif np.issubdtype(type(data), np.floating):
        return

    elif isinstance(data, pd.Series):
        index_name = data.index.name or "index"
        value_name = data.name or "value"
        _ = [
            {
                index_name: ensure_json_serializable(idx),
                value_name: ensure_json_serializable(val),
            }
            for idx, val in data.iteritems()
        ]
        return
    elif isinstance(data, pd.DataFrame):
        return ensure_json_serializable(data.to_dict(orient="records"))

    elif isinstance(data, decimal.Decimal):
        return

    else:
        raise Exception(
            f"{data} is of type {type(data).__name__} which cannot be serialized to json"
        )


############################################################################################
# ValidationResult
############################################################################################


class ValidationResult:
    def __init__(self, success=None, result=None, exception_info=None):
        self.success = success
        self.result = result or {}

        self.exception_info = exception_info or {
            "raised_exception": False,
            "where": None,
            "why": None,
        }

    def __repr__(self):
        return json.dumps(self.to_json_dict(), indent=2)

    def __str__(self):
        return json.dumps(self.to_json_dict(), indent=2)

    def to_json_dict(self):
        return {
            "success": self.success,
            "result": convert_to_json_serializable(self.result),
            "exception_info": convert_to_json_serializable(self.exception_info),
        }


############################################################################################
# Validation checks
############################################################################################


def check_column_exists(
    data: pd.DataFrame,
    column: str,
    column_index: int = None,
    catch_exceptions: bool = None,
):
    """Expect the specified column to exist.

    Args:
        data (pandas DataFrame): The data to validate.
        column (str): The column name.
        column_index (int or None): If not None, checks also for the location column_index
            (zero-indexed).
        catch_exceptions (boolean or None): If True, then catch exceptions and include them
            as part of the result object.

    Returns:
        ValidationResult
    """
    exception_info = None

    try:
        if column_index is not None:
            success = data.columns[column_index] == column
        else:
            success = column in data.columns

    except Exception as ex:
        if catch_exceptions:
            exceptiondata = traceback.format_exc().splitlines()
            exception_info = {
                "raised_exception": True,
                "where": exceptiondata[1],
                "why": exceptiondata[-1],
            }
            return ValidationResult(
                success=False, result=None, exception_info=exception_info
            )
        else:
            raise ex

    else:
        return ValidationResult(
            success=bool(success), result=None, exception_info=exception_info
        )


def check_column_values_unique(
    data: pd.DataFrame,
    column: str,
    catch_exceptions: bool = None,
):
    """Expect the specified column's values to be unique.

    Args:
        data (pandas DataFrame): The data to validate.
        column (str): The column name.
        catch_exceptions (boolean or None): If True, then catch exceptions and include them
            as part of the result object.

    Returns:
        ValidationResult
    """
    exception_info = None

    try:
        n_obs = len(data[column])
        n_missing = data[column].isna().sum()
        count = n_obs - data[column].nunique()
        success = count == 0

    except Exception as ex:
        if catch_exceptions:
            exceptiondata = traceback.format_exc().splitlines()
            exception_info = {
                "raised_exception": True,
                "where": exceptiondata[1],
                "why": exceptiondata[-1],
            }
            return ValidationResult(
                success=False, result=None, exception_info=exception_info
            )
        else:
            raise ex

    else:
        result = {
            "element_count": n_obs,
            "missing_count": n_missing,
            "unexpected_count": count,
        }
        return ValidationResult(
            success=bool(success),
            result=result,
            exception_info=exception_info,
        )


def check_column_values_increasing(
    data: pd.DataFrame,
    column: str,
    parse_as_datetimes: bool = False,
    catch_exceptions: bool = None,
):
    """Expect column values to be increasing.

    Args:
        data (pandas DataFrame): The data to validate.
        column (str): The column name.
        parse_as_datetimes (boolean or None) : If True, column values will be parsed to
            datetimes before making comparisons
        catch_exceptions (boolean or None): If True, then catch exceptions and include them
            as part of the result object.

    Returns:
        ValidationResult
    """
    exception_info = None

    try:
        expected_column = data[column]

        if parse_as_datetimes:
            if data[column].dtype == np.int64:
                expected_column = pd.to_datetime(data[column].astype(str))
            else:
                expected_column = pd.to_datetime(data[column])
        success = expected_column.is_monotonic_increasing

    except Exception as ex:
        if catch_exceptions:
            exceptiondata = traceback.format_exc().splitlines()
            exception_info = {
                "raised_exception": True,
                "where": exceptiondata[1],
                "why": exceptiondata[-1],
            }
            return ValidationResult(
                success=False, result=None, exception_info=exception_info
            )
        else:
            raise ex

    else:
        return ValidationResult(
            success=bool(success), result=None, exception_info=exception_info
        )


def check_column_type_datetime(
    data: pd.DataFrame, column: str, catch_exceptions: bool = None
):
    """Expect column values to be of datetime type.

    Args:
        data (pandas DataFrame): The data to validate.
        column (str): The column name.
        catch_exceptions (boolean or None): If True, then catch exceptions and include them
            as part of the result object.

    Returns:
        ValidationResult
    """
    exception_info = None

    try:
        success = is_datetime(data[column])

    except Exception as ex:
        if catch_exceptions:
            exceptiondata = traceback.format_exc().splitlines()
            exception_info = {
                "raised_exception": True,
                "where": exceptiondata[1],
                "why": exceptiondata[-1],
            }
            return ValidationResult(
                success=False, result=None, exception_info=exception_info
            )
        else:
            raise ex

    else:
        return ValidationResult(
            success=bool(success), result=None, exception_info=exception_info
        )


def check_column_values_not_null(
    data: pd.DataFrame,
    column: str,
    catch_exceptions: bool = None,
):
    """Expect column values to not be null.

    Args:
        data (pandas DataFrame): The data to validate.
        column (str): The column name.
        catch_exceptions (boolean or None): If True, then catch exceptions and include them
            as part of the result object.

    Returns:
        ValidationResult
    """
    exception_info = None
    expected_column = data[column]

    try:
        n_obs = len(expected_column)
        count = expected_column.isna().sum()

    except Exception as ex:
        if catch_exceptions:
            exceptiondata = traceback.format_exc().splitlines()
            exception_info = {
                "raised_exception": True,
                "where": exceptiondata[1],
                "why": exceptiondata[-1],
            }
            return ValidationResult(
                success=False, result=None, exception_info=exception_info
            )
        else:
            raise ex

    else:
        success = count == 0
        result = {
            "element_count": n_obs,
            "missing_count": count,
            "unexpected_count": count,
        }
        return ValidationResult(
            success=bool(success),
            result=result,
            exception_info=exception_info,
        )


############################################################################################
# Validation corrective actions
############################################################################################


def remove_duplicate_dates(
    data: pd.DataFrame, date_column: str, threshold: float = 0.25
):
    """Remove duplicate timestamps.

    Args:
        data (pandas.DataFrame): The data to correct.
        date_column (str): The name of the column that contains datetime information.
        threshold (float, optional): If the range of the values that share a timestamp is less
            than `threshold` times the standard deviation of the data, they are replaced by their
            average. Otherwise, they are treated as missing values. Defaults to 0.25.

    Returns:
        pandas.DataFrame: The input data without the duplicate timestamps.
    """

    results = []
    if not is_datetime(data[date_column]):
        data[date_column] = pd.to_datetime(data[date_column])

    data = data.set_index(date_column)

    for col in data.columns:
        data_ = data[[col]]
        dupl_range = (
            data_[data_.index.duplicated(keep=False)]
            .groupby([lambda x: x])
            .agg([lambda x: x.max() - x.min(), lambda x: x.mean()])
        )
        dupl_range.columns = dupl_range.columns.droplevel(1)
        dupl_range.columns = ["range", "mean"]

        data_ = data_[~data_.index.duplicated(keep="first")]
        data_.loc[data_.index.isin(dupl_range.index)] = np.nan

        fill_index = dupl_range[
            dupl_range["range"] < threshold * np.std(data_).item()
        ].index
        fill_values = dupl_range.loc[fill_index]["mean"].values
        data_.loc[data_.index.isin(fill_index)] = fill_values.reshape(-1, 1)
        results.append(data_)

    return pd.concat(results, axis=1).reset_index()


def apply_rebind_names(data: pd.DataFrame, rebind_names: Dict[str, str]):
    """Rename the input dataframe columns.

    Args:
        data (pandas.DataFrame): Input dataset.
        rebind_names (dict): Dictionary of key/value pairs used to redefine
            column names for the input dataset.

    Returns:
        pandas.DataFrame: The input dataframe with renamed columns.
    """
    if rebind_names and any(rebind_names.values()):
        columns = {}
        for key, val in rebind_names.items():
            if (val is not None) and (val in data.columns):
                columns[val] = key
        data = data.rename(columns=columns)

    if np.any(data.columns.duplicated()):
        raise ValueError(
            "Duplicate column names found: "
            f"{data.columns[data.columns.duplicated()].tolist()}"
        )

    return data


############################################################################################
# Composite functions
############################################################################################


def validate_dataset(
    data: pd.DataFrame,
    rebind_names: Dict[str, str] = None,
    date_format: str = "%Y-%m-%d %H:%M:%S",
    threshold: float = 0.25,
) -> pd.DataFrame:
    """Validate a dataset.

    Args:
        data (pandas.DataFrame): The data to validate.
        rebind_names (dict, optional): Dictionary of key/value pairs used to redefine column
            names for the input dataset. Defaults to None.
        date_format (str, optional): The strftime to parse the timestamp dates. Defaults to
            %Y-%m-%d %H:%M:%S.
        threshold (float, optional): If the range of the values that share a timestamp is less
            than `threshold` times the standard deviation of the data, they are replaced by their
            average. Otherwise, they are treated as missing values. Defaults to 0.25.

    Returns:
        pandas.DataFrame: The validated dataframe.
    """
    if rebind_names:
        data = apply_rebind_names(data, rebind_names)

    if not check_column_exists(data, "timestamp").success:
        raise ValueError(f"Column `timestamp` is missing from input dataset")

    if not check_column_type_datetime(data, "timestamp").success:
        try:
            data["timestamp"] = pd.to_datetime(data["timestamp"], format=date_format)
            data = data.dropna(subset="timestamp")
        except ValueError:
            raise ValueError(f"Column `timestamp` must be in datetime format")

    if not check_column_values_increasing(data, "timestamp").success:
        data = data.sort_values(by=["timestamp"])

    if not check_column_values_unique(data, "timestamp").success:
        data = remove_duplicate_dates(data, "timestamp", threshold=threshold)

    data = data.set_index("timestamp")
    to_drop = data.filter(like="Unnamed", axis=1).columns
    if len(to_drop) > 0:
        data = data.drop(to_drop, axis=1)

    return data



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import datetime
import logging
from collections import defaultdict
from typing import Callable, Dict, Literal, Union

import numpy as np
import pandas as pd
from pandas.api.types import is_float_dtype as is_float

from .validation import validate_dataset

logger = logging.getLogger("eensight")


def get_hourly_timesteps(index):
    """Get the time steps in the provided index with duration less than one day.

    Args:
        index (pd.DatetimeIndex): The input index.

    Returns:
        pd.TimedeltaIndex: The identified time steps.
    """

    if isinstance(index, pd.DatetimeIndex):
        index = index.to_series()

    time_steps = np.array(
        [pd.Timedelta(x) for x in index.diff().unique() if pd.notnull(x)]
    )
    return time_steps[time_steps < pd.Timedelta(days=1)]


def get_time_step(index: pd.DatetimeIndex):
    """Get the most frequent time step of the provided index.

    Args:
        index (pd.DatetimeIndex): The input index.

    Returns:
        pd.Timedelta: The identified time step.
    """
    dt = index.to_series().diff()
    time_step = dt.iloc[dt.values.nonzero()[0]].mode()[0]
    return time_step if (time_step < pd.Timedelta(days=1)) else pd.Timedelta(days=1)


def _align_by_distance(data, index, tolerance):
    data = pd.merge_asof(
        pd.DataFrame(index=index),
        data,
        left_index=True,
        right_index=True,
        tolerance=tolerance,
        direction="nearest",
    )
    return data


def _align_by_value(data, index, tolerance):
    full_index = (
        pd.concat((index.to_series(), data.index.to_series()), ignore_index=True)
        .sort_values()
        .drop_duplicates()
    )

    time_step = pd.Timedelta(
        minutes=np.gcd.reduce(
            [int(delta.seconds / 60) for delta in get_hourly_timesteps(full_index)]
        )
    )

    full_index = pd.date_range(
        start=datetime.datetime.combine(full_index.min().date(), datetime.time(0, 0)),
        end=datetime.datetime.combine(
            full_index.max().date() + datetime.timedelta(days=1), datetime.time(0, 0)
        ),
        freq=time_step,
        inclusive="left",
    )

    data = data.reindex(full_index)
    data = data.interpolate(
        method="slinear",
        limit=int(tolerance / time_step),
        limit_area="inside",
        limit_direction="both",
        axis=0,
    ).reindex(index)

    return data


def align_to_index(
    data: pd.DataFrame,
    index: pd.DatetimeIndex,
    mode: Union[
        Literal["value", "distance"], Dict[str, Literal["value", "distance"]]
    ] = None,
    tolerance: Union[pd.Timedelta, Dict[str, pd.Timedelta]] = None,
    cumulative: Union[bool, Dict[str, bool]] = False,
) -> pd.DataFrame:
    """Aling the index of the input dataframe to the provided index.

    Args:
        data (pandas.DataFrame): The dataframe to align. The dataframe's index
            must be a pandas DatetimeIndex and in increasing order.
        index (pandas.DatetimeIndex): The index with which the input dataframe will be
            aligned.
        mode ({'value', 'distance'} or dict, optional): Defines how index alignment should
            be carried out. `mode='value'` interpolates the data to match the primary index,
            while `mode='distance'` matches the primary index on the nearest key of the feature's
            index. If the input dataframe has more than one columns, `mode` can be a dictionary
            that maps feature names to `mode` values. Defaults to None.
        tolerance (pandas.Timedelta or dict, optional): If `mode='value'`, `tolerance` dictates
            how far the interpolation can reach. As an example, outdoor temperature changes slowly,
            so interpolating 2 or 3 hours ahead should not be a problem. If `mode='distance'`,
            `tolerance` is the maximum time distance to match a timestamp of the input dataframe
            with a timestamp in the primary index. If the input dataframe has more than one
            columns, `tolerance` can be a dictionary that maps feature names to tolerance
            values. Defaults to None.
        cumulative (bool or dict, optional): If (a) `cumulative=True` and (b) the index of the input
            dataframe and the `index` are misaligned, and (c) the data is numerical, the data will
            be aligned after `cumsum` has been applied on it. If the input dataframe has more than
            one columns, `cumulative` can be a dictionary that maps feature names to cumulative
            values. Defaults to False.

    Returns:
        pandas.DataFrame: The input dataframe with aligned index.

    Note:
        If a feature that is both numerical and cumulative must be aligned, it will be normalized
        by dividing its value at any given timestamp by the number of minutes between this timestamp
        and the immediately preceding one (so that to avoid misleading values when the primary index
        has time gaps).
    """
    if (not isinstance(data.index, pd.DatetimeIndex)) or (
        not isinstance(index, pd.DatetimeIndex)
    ):
        raise ValueError(
            "Both the input dataframe's index and the `index` to align to must be"
            " pandas DatetimeIndex."
        )

    if (not data.index.is_monotonic_increasing) or (not index.is_monotonic_increasing):
        raise ValueError(
            "Both the input dataframe's index and the `index` to align to must be "
            "in increasing order."
        )

    data_time_step = get_time_step(data.index)
    index_time_step = get_time_step(index)

    if not isinstance(mode, dict):
        mode = {col: mode for col in data.columns}
    if not isinstance(tolerance, dict):
        tolerance = {col: tolerance for col in data.columns}
    if not isinstance(cumulative, dict):
        cumulative = {col: cumulative for col in data.columns}

    if data_time_step == index_time_step:  # aligned
        return pd.DataFrame(index=index).join(data, how="left")

    elif (
        (data_time_step < pd.Timedelta(days=1))
        and (index_time_step < pd.Timedelta(days=1))
    ) or (
        (data_time_step >= pd.Timedelta(days=1))
        and (index_time_step >= pd.Timedelta(days=1))
    ):
        to_merge = []
        for col in data.columns:
            tolerance_col = tolerance.get(col, None)
            mode_col = mode.get(col, None)

            if tolerance_col and mode_col:
                if is_float(data[col]) and cumulative.get(col, False):
                    aligned = (
                        _align_by_value(data[col].cumsum(), index, tolerance_col).diff()
                        if mode[col] == "value"
                        else _align_by_distance(
                            data[col].cumsum(), index, tolerance_col
                        ).diff()
                    )
                    aligned = aligned.div(
                        aligned.index.to_series()
                        .diff()
                        .map(lambda x: x.total_seconds() / 60),
                        axis=0,
                    )
                    logger.info(
                        f"Cumulative feature {col} normalized by time step length."
                    )
                    to_merge.append(aligned)

                elif is_float(data[col]):
                    aligned = (
                        _align_by_value(data[col], index, tolerance_col)
                        if mode[col] == "value"
                        else _align_by_distance(data[col], index, tolerance_col)
                    )
                    to_merge.append(aligned)

                else:
                    aligned = _align_by_distance(data[col], index, tolerance_col)
                    to_merge.append(aligned)

            else:
                aligned = pd.DataFrame(index=index).join(data[[col]], how="left")
                to_merge.append(aligned)

        return pd.concat(to_merge, axis=1)

    elif index_time_step < pd.Timedelta(days=1):
        full_index = (
            pd.concat((index.to_series(), data.index.to_series()), ignore_index=True)
            .sort_values()
            .drop_duplicates()
        )
        data = data.reindex(full_index).loc[full_index.min() : full_index.max()]
        for _, grouped in data.groupby(pd.Grouper(freq=data_time_step)):
            data.loc[grouped.index] = grouped.fillna(method="ffill").fillna(
                method="bfill"
            )
        return data.reindex(index)

    else:
        raise NotImplementedError(
            "Aligning sub-daily data to a daily index is not supported."
        )


def merge_data_list(
    data_list: list,
    primary: pd.DatetimeIndex,
    mode: Dict[str, Literal["value", "distance"]] = None,
    tolerance: Dict[str, pd.Timedelta] = None,
    cumulative: Dict[str, bool] = False,
):
    """Merge together all dataframes in the provided list.

    The dataframes should not have column names in common.

    Args:
        data_list (list of pandas DataFrames): The list of dataframes to merge.
        primary (pandas.DatetimeIndex): The shared index (i.e. the index to align all
            partitioned datasests with).
        mode (dict, optional): Dictionary that maps feature names to how index alignment
            should be carried out for them. 'value' interpolates the data to match the
            primary index, while 'distance' matches the primary index on the nearest key
            of the feature's index. It is needed when the index of one or more of the
            dataframes and the `primary` index are misaligned. Defaults to None.
        tolerance (dict, optional): Dictionary that maps feature names to: (a) how far
            interpolation can reach if the corresponing mode is 'value' or (b) the maximum
            time distance to match a timestamp of the input dataframe with a timestamp in
            the primary index if the corresponing mode is 'distance'. It is needed when the
            index of one or more of the dataframes and the `primary` index are misaligned.
            Defaults to None.
        cumulative (dict, optional): Dictionary that maps feature names to boolean values
            indicating whether the data is cumulative or not. Defaults to False.

    Returns:
        pandas.DataFrame: The merged dataframe.
    """

    result = None
    for data in data_list:
        result = pd.concat(
            [
                result,
                align_to_index(
                    data, primary, mode=mode, tolerance=tolerance, cumulative=cumulative
                ),
            ],
            axis=1,
        )
    return result


def validate_partitions(
    partitioned_data: Dict[str, Callable],
    rebind_names: Dict[str, str] = None,
    date_format: str = "%Y-%m-%d %H:%M:%S",
    threshold: float = 0.25,
    primary: pd.DatetimeIndex = None,
    mode: Dict[str, Literal["value", "distance"]] = None,
    tolerance: Dict[str, pd.Timedelta] = None,
    cumulative: Dict[str, bool] = False,
) -> pd.DataFrame:
    """Validate and merge input data partitions into one pandas DataFrame.

    Args:
        partitioned_data: A dictionary with file names as keys and load functions as
            values. Each load function returns a pandas DataFrame.
        rebind_names (dict, optional): Dictionary of key/value pairs used to redefine
            column names for the input dataset. Defaults to None.
        date_format (str, optional): The strftime to parse the timestamp dates. Defaults
            to %Y-%m-%d %H:%M:%S.
        threshold (float, optional): If the range of the values that share a timestamp
            is less than `threshold` times the standard deviation of the data, they are
            replaced by their average. Otherwise, they are treated as missing values.
            Defaults to 0.25.
        primary (pandas.DatetimeIndex, optional): The shared index (i.e. the index to
            align all partitioned datasests with). It is required if the partitions
            include dataframes of different features (instead of only different time
            periods of one feature). Defaults to None.
        mode (dict, optional): Dictionary that maps feature names to how index alignment
            should be carried out for them. 'value' interpolates the data to match the
            primary index, while 'distance' matches the primary index on the nearest key
            of the feature's index. It is needed when the index of one or more of the
            dataframes and the `primary` index are misaligned. If `primary` is None, it
            will be ignored.Defaults to None.
        tolerance (dict, optional): Dictionary that maps feature names to: (a) how far
            interpolation can reach if the corresponing mode is 'value' or (b) the maximum
            time distance to match a timestamp of the input dataframe with a timestamp in
            the primary index if the corresponing mode is 'distance'. It is needed when the
            index of one or more of the dataframes and the `primary` index are misaligned.
            If `primary` is None, it will be ignored. Defaults to None.
        cumulative (dict, optional): Dictionary that maps feature names to boolean values
            indicating whether the data is cumulative or not. If `primary` is None, it will
            be ignored. Defaults to False.

    Returns:
        pandas.DataFrame: Dataframe with all loaded partitions merged.
    """
    datasets = defaultdict(list)
    for _, load_func in partitioned_data.items():
        df = load_func()
        df = validate_dataset(
            df,
            rebind_names=rebind_names,
            date_format=date_format,
            threshold=threshold,
        )
        datasets[tuple(set(df.columns))].append(df)

    if (len(datasets) > 1) and (primary is None):
        raise ValueError(
            "`primary` index is required if the partitions "
            "include dataframes of different features."
        )

    horizontal_data = []
    for items in datasets.values():
        vertical_data = pd.concat(items, axis=0)
        if vertical_data.index.has_duplicates:
            raise ValueError(
                "Partitions contain overlapping datetime indices. "
                "This leads to duplicate index entries when merging."
            )
        horizontal_data.append(vertical_data.sort_index())

    if len(horizontal_data) == 1:
        if primary is not None:
            return align_to_index(
                horizontal_data[0],
                primary,
                mode=mode,
                tolerance=tolerance,
                cumulative=cumulative,
            )
        return horizontal_data[0]

    return merge_data_list(
        horizontal_data, primary, mode=mode, tolerance=tolerance, cumulative=cumulative
    )



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

from kedro.pipeline import Pipeline, node, pipeline

from .nodes import adjust_activity_levels

ppl_nodes = {
    "adjust_activity_levels": node(
        func=adjust_activity_levels,
        inputs=[
            "preprocessed-features",
            "preprocessed-labels",
            "model-autoenc",
            "params:activity.non_occ_features",
            "params:activity.cat_features",
            "params:activity.assume_hurdle",
            "params:activity.n_trials_adjust",
            "params:activity.upper_bound",
            "params:activity.verbose",
        ],
        outputs="activity-adjusted",
        name="adjust_activity_levels",
    ),
}

adjust_apply = pipeline(
    pipeline([ppl_nodes["adjust_activity_levels"]]),
    inputs=["model-autoenc"],
    parameters=[
        "activity.non_occ_features",
        "activity.cat_features",
        "activity.assume_hurdle",
        "activity.n_trials_adjust",
        "activity.upper_bound",
        "activity.verbose",
    ],
    namespace="apply",
)


def create_pipeline(**kwargs):
    return Pipeline([adjust_apply], tags=["apply_autoenc"])



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
from typing import Any, Union

import numpy as np
import pandas as pd
import scipy
from kedro.framework.session import KedroSession
from pandas.api.types import is_bool_dtype as is_bool
from pandas.api.types import is_categorical_dtype as is_category
from pandas.api.types import is_datetime64_any_dtype as is_datetime
from pandas.api.types import is_integer_dtype as is_integer
from pandas.api.types import is_object_dtype as is_object
from sklearn.utils.validation import column_or_1d

from eensight.framework.startup import bootstrap_project
from eensight.settings import PROJECT_PATH


def as_list(val: Any) -> list:
    """Cast input as list.

    Helper function, always returns a list of the input value.
    """
    if isinstance(val, str):
        return [val]
    if hasattr(val, "__iter__"):
        return list(val)
    if val is None:
        return []
    return [val]


def as_series(x: Union[np.ndarray, pd.Series, pd.DataFrame]):
    """Cast an iterable to a Pandas Series object."""
    if isinstance(x, pd.Series):
        return x
    if isinstance(x, pd.DataFrame):
        return x.iloc[:, 0]
    else:
        return pd.Series(column_or_1d(x))


def get_categorical_cols(X: pd.DataFrame, int_is_categorical: bool = False) -> list:
    """Return the names of the categorical columns in the input DataFrame.

    Args:
        X (pandas.DataFrame): Input dataframe.
        int_is_categorical (bool, optional): If True, integer types are
            considered categorical. Defaults to False.

    Returns:
        list: The names of categorical columns in the input DataFrame.
    """
    obj_cols = []
    for col in X.columns:
        # check if it is date
        if is_datetime(X[col]):
            continue
        # check if it is bool, object or category
        if is_bool(X[col]) or is_object(X[col]) or is_category(X[col]):
            obj_cols.append(col)
            continue
        # check if it is integer
        if int_is_categorical and is_integer(X[col]):
            obj_cols.append(col)
            continue
    return obj_cols


def tensor_product(a: np.ndarray, b: np.ndarray, reshape=True):
    """Compute the tensor product of two matrices.

    Args:
        a (numpy array of shape (n, m_a)): The first matrix.
        b (numpy array of shape (n, m_b)): The second matrix.
        reshape (bool, optional): Whether to reshape the result to be 2D (n, m_a * m_b)
            or return a 3D tensor (n, m_a, m_b). Defaults to True.

    Raises:
        ValueError: If input arrays are not 2-dimensional.
        ValueError: If both input arrays do not have the same number of samples.

    Returns:
        numpy.ndarray of shape (n, m_a * m_b) if `reshape = True` else of shape (n, m_a, m_b).
    """
    if (a.ndim != 2) or (b.ndim != 2):
        raise ValueError("Inputs must be 2-dimensional")

    na, ma = a.shape
    nb, mb = b.shape

    if na != nb:
        raise ValueError("Both arguments must have the same number of samples")

    if scipy.sparse.issparse(a):
        a = a.A
    if scipy.sparse.issparse(b):
        b = b.A

    product = a[..., :, None] * b[..., None, :]
    if reshape:
        return product.reshape(na, ma * mb)
    return product


def load_catalog(store_uri: str, site_id: str, namespace: str, run_id: str = None):
    """Load the data catalog for the provided building id and namespace.

    Args:
        store_uri (str): The URI where the generated data and models should be stored.
            It is expected to be a local path. In addition, the function will use
            `{store_uri}/{site_id}/01_raw/{namespace}/features` path as default for the
            input features and `{store_uri}/{site_id}/01_raw/{namespace}/labels` for the
            input labels.
        site_id (str): The id of the site/building to load (must be one of
            the datasets already provided by `eensight`).
        namespace (str): The namespace for which to load data (`train`, `test`
            or `apply`). Not all datasets already provided by `eensight` have
            data for all namespaces.
        run_id (str, optional): The ID string for the run that contains the artifacts
            to be used. If not provided, the untracked artifacts will be used.

    Returns:
        kedro.io.DataCatalog: The data catalog.
    """

    extra_params = {
        "app": {
            "store_uri": os.path.abspath(store_uri),
            "site_id": site_id,
            "namespace": namespace,
            "run_id": run_id,
            "features": {
                "format": "csv",
                "load_args": {},
            },
            "labels": {"format": "csv", "load_args": {}},
        }
    }

    bootstrap_project(PROJECT_PATH)
    with KedroSession.create(
        package_name="eensight",
        project_path=PROJECT_PATH,
        extra_params=extra_params,
        save_on_close=False,
    ) as session:
        catalog = session.load_context().catalog
        catalog._logger.setLevel(logging.WARNING)
        return catalog



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import List, Union

import pandas as pd

from eensight.methods.prediction.activity import estimate_activity
from eensight.methods.prediction.baseline import UsagePredictor
from eensight.methods.prediction.metrics import cvrmse, nmbe

logger = logging.getLogger("eensight")


def create_activity_feature(
    features: pd.DataFrame,
    labels: pd.DataFrame,
    non_occ_features: Union[str, List[str]],
    cat_features: Union[str, List[str]],
    exog: str,
    assume_hurdle: bool,
    n_trials: int,
    verbose: bool,
    adjusted_activity: pd.DataFrame,
):
    """Estimate activity levels.

    Args:
        features (pd.DataFrame): The feature data.
        labels (pd.DataFrame): The target data.
        non_occ_features (Union[str, List[str]]): The names of the features to
            use to fit the two quantile regression models on the label data. If
            not provided, "temperature" will be used.
        cat_features (Union[str, List[str]]): The name(s) of the categorical
            features in the input data.
        exog (str): The name of the feature to use for determining how to best
            split the data into two regimes. It should not contain categorical data.
        assume_hurdle (bool): If True, activity estimation will assume that energy
            consumption has been generated by a hurdle model.
        n_trials (int): The number of iterations for the underlying optimization.
        verbose (bool): If True, a progress bar is visible.
        adjusted_activity (pd.DataFrame): Placeholder to pass adjusted activity
            levels if they exist.

    Returns:
        pandas.DataFrame: The estimated activity time series.
    """

    if isinstance(adjusted_activity, pd.DataFrame):
        logger.warning(
            "Skipping activity estimation because adjusted activity was found"
        )
        return adjusted_activity

    activity, best_params = estimate_activity(
        features,
        labels,
        non_occ_features=non_occ_features,
        cat_features=cat_features,
        exog=exog,
        assume_hurdle=assume_hurdle,
        n_trials=n_trials,
        verbose=verbose,
        return_params=True,
    )

    logger.info(f"Estimated parameters: {best_params}")
    return activity.to_frame("activity")


def fit_predict(
    features: pd.DataFrame,
    labels: pd.DataFrame,
    lags: dict,
    cat_features: Union[str, List[str]],
    validation_size: float,
):
    """Fit a predictive model and apply it on the training data.

    Args:
        features (pd.DataFrame): The feature data.
        labels (pd.DataFrame): The target data.
        lags (dict): Dictionary with keys that correspond to feature names,
            and values that are lists containing the time lags of that
            feature to add as additional regressors.
        cat_features (str or list of str): The names of the categorical features
            in the input data.
        validation_size (float): The size of the validation dataset (as
            percentage of all training data) to use for identifying optimal
            number of iterations for the Gradient Boosting model.

    Returns:
        tuple containing

        - **model** (*UsagePredictor*): The fitted model.
        - **prediction** (*pandas.DataFrame*): The in-sample prediction.
        - **metrics** (*dict*): The in-sample CV(RMSE) and NMBE metrics.
    """

    model = UsagePredictor(
        lags=lags,
        cat_features=cat_features,
        validation_size=validation_size,
        skip_calendar=False,
    )
    model = model.fit(features, labels)
    logger.info(f"Model scores: {model.scores_}")

    prediction = model.predict(features).to_frame("consumption")

    metrics = {
        "CVRMSE": cvrmse(labels, prediction),
        "NMBE": nmbe(labels, prediction),
    }
    logger.info(f"In-sample metrics: {metrics}")

    return model, prediction, metrics


def predict(features: pd.DataFrame, model: UsagePredictor):
    """Predict energy consumption given input data.

    Args:
        features (pd.DataFrame): The feature data.
        model (UsagePredictor): The fitted model.

    Returns:
        pandas.DataFrame: The prediction.
    """
    prediction = model.predict(features).to_frame("consumption")
    return prediction


def fit_predict_autoenc(
    features: pd.DataFrame,
    labels: pd.DataFrame,
    activity: pd.DataFrame,
    lags: dict,
    cat_features: Union[str, List[str]],
    validation_size: float,
    assume_hurdle: bool,
):
    """Fit an autoencoder model and apply it on the training data.

    Args:
        features (pd.DataFrame): The feature data.
        labels (pd.DataFrame): The target data.
        activity (pd.DataFrame): The estimated activity levels.
        lags (dict): Dictionary with keys that correspond to feature names,
            and values that are lists containing the time lags of that
            feature to add as additional regressors.
        cat_features (str or list of str): The names of the categorical features
            in the input data.
        validation_size (float): The size of the validation dataset (as
            percentage of all training data) to use for identifying optimal
            number of iterations for the Gradient Boosting model.
        assume_hurdle (bool): If True, activity estimation will assume that energy
            consumption has been generated by a hurdle model.

    Returns:
        tuple containing

        - **model** (*UsagePredictor*): The fitted model.
        - **prediction** (*pandas.DataFrame*): The in-sample prediction.
        - **metrics** (*dict*): The in-sample CV(RMSE) and NMBE metrics.
    """

    if assume_hurdle:
        selected = activity["activity"] == 1
        X_selected = features[selected]
        y_selected = labels[selected]

        model = UsagePredictor(
            lags=lags,
            cat_features=cat_features,
            validation_size=validation_size,
            skip_calendar=False,
        )
        model = model.fit(X_selected, y_selected)
        logger.info(f"Model scores: {model.scores_}")

        prediction = model.predict(features).to_frame("consumption")
        metrics = {
            "CVRMSE": cvrmse(y_selected, prediction[selected]),
            "NMBE": nmbe(y_selected, prediction[selected]),
        }
        logger.info(f"In-sample metrics: {metrics}")

    else:
        X_act = pd.concat([features, activity], axis=1)
        model = UsagePredictor(
            lags=lags,
            cat_features=cat_features,
            validation_size=validation_size,
            skip_calendar=True,
        )
        model = model.fit(X_act, labels)
        logger.info(f"Model scores: {model.scores_}")

        prediction = model.predict(X_act).to_frame("consumption")

        metrics = {
            "CVRMSE": cvrmse(labels, prediction),
            "NMBE": nmbe(labels, prediction),
        }
        logger.info(f"In-sample metrics: {metrics}")

    return model, prediction, metrics


def predict_autoenc(
    features: pd.DataFrame,
    activity: pd.DataFrame,
    model: UsagePredictor,
    assume_hurdle: bool,
):
    """Predict energy consumption given input data.

    Args:
        features (pd.DataFrame): The feature data.
        activity (pd.DataFrame): The estimated activity levels.
        model (UsagePredictor): The fitted model.
        assume_hurdle (bool): If True, activity estimation will assume that energy
            consumption has been generated by a hurdle model.

    Returns:
        pandas.DataFrame: The prediction.
    """

    if assume_hurdle:
        prediction = model.predict(features).to_frame("consumption")

    else:
        X_act = pd.concat([features, activity], axis=1)
        prediction = model.predict(X_act).to_frame("consumption")

    return prediction



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import logging
from glob import iglob
from pathlib import Path
from typing import AbstractSet, Any, Dict, Iterable, List, Optional, Set

import jinja2
from kedro.config import (
    AbstractConfigLoader,
    BadConfigException,
    MissingConfigException,
)
from omegaconf import OmegaConf

SUPPORTED_EXTENSIONS = [".yml", ".yaml"]

logger = logging.getLogger("ConfigLoader")


########################################################################################
# Utility functions
# From https://github.com/kedro-org/kedro/blob/0.18.1/kedro/config/common.py
########################################################################################


def _remove_duplicates(items: Iterable[Any]):
    """Remove duplicates while preserving the order."""
    unique_items = []  # type: List[str]
    for item in items:
        if item not in unique_items:
            unique_items.append(item)
        else:
            logger.warning(
                f"Duplicate environment detected! "
                f"Skipping re-loading from configuration path: {item}"
            )
    return unique_items


def _check_duplicate_keys(
    processed_files: Dict[Path, AbstractSet[str]], filepath: Path, conf: Dict[str, Any]
) -> None:
    duplicates = []

    for processed_file, keys in processed_files.items():
        overlapping_keys = conf.keys() & keys

        if overlapping_keys:
            sorted_keys = ", ".join(sorted(overlapping_keys))
            if len(sorted_keys) > 100:
                sorted_keys = sorted_keys[:100] + "..."
            duplicates.append(f"{processed_file}: {sorted_keys}")

    if duplicates:
        dup_str = "\n- ".join(duplicates)
        raise ValueError(f"Duplicate keys found in {filepath} and:\n- {dup_str}")


def _path_lookup(conf_path: Path, patterns: Iterable[str]) -> Set[Path]:
    """Return a set of all configuration files from ``conf_path`` or
    its subdirectories, which satisfy a given list of glob patterns.
    """
    config_files = set()
    conf_path = conf_path.resolve()

    for pattern in patterns:
        # `Path.glob()` ignores the files if pattern ends with "**",
        # therefore iglob is used instead
        for each in iglob(str(conf_path / pattern), recursive=True):
            path = Path(each).resolve()
            if path.is_file() and path.suffix in SUPPORTED_EXTENSIONS:
                config_files.add(path)

    return config_files


def _lookup_config_filepaths(
    conf_path: Path,
    patterns: Iterable[str],
    processed_files: Set[Path],
) -> List[Path]:
    config_files = _path_lookup(conf_path, patterns)

    seen_files = config_files & processed_files
    if seen_files:
        logger.warning(
            "Config file(s): %s already processed, skipping loading...",
            ", ".join(str(seen) for seen in sorted(seen_files)),
        )
        config_files -= seen_files

    return sorted(config_files)


##########################################################################################
# ConfigLoader
##########################################################################################


class ConfigLoader(AbstractConfigLoader):
    """Recursively scan directories (config paths) contained in ``conf_source`` for
    configuration files with a ``yaml`` or ``yml`` extension, load them, and return them
    in the form of a config dictionary.

    The first processed config path is the ``base`` directory inside ``conf_source``. The
    optional ``env`` argument can be used to specify a subdirectory of ``conf_source`` to
    process as a config path after ``base``. When the same top-level key appears in any 2
    config files located in the same (sub)directory, a ``ValueError`` is raised. When the
    same key appears in any 2 config files located in different (sub)directories, the last
    processed config path takes precedence and overrides this key.

    Args:
        conf_source (str): Path to use as root directory for loading configuration.
        env (str, optional): Environment that will take precedence over `base_env`. Defaults
            to None.
        runtime_params (dict, optional): Extra parameters passed to a Kedro run. Defaults to
            None.
        config_patterns: Regex patterns that specify the naming convention for configuration
            files so they can be loaded. Can be customised by supplying config_patterns as
            in `CONFIG_LOADER_ARGS` in `settings.py`.
        base_env (str, optional): Keyword-only argument for the name of the base environment.
            This is used in the `conf_paths` property method to construct the configuration
            paths. Defaults to "base".
        default_run_env (str, optional): Keyword-only argument for the name of the default run
            environment. This is used in the `conf_paths` property method to construct the
            configuration paths. Can be overriden by supplying the `env` argument. Defaults to
            "local".
        globals_pattern (str, optional): Keyword-only argument specifying a glob pattern. Files
            that match this pattern will be loaded as files containing global values for variable
            interpolation. Defaults to None.
    """

    def __init__(
        self,
        conf_source: str,
        env: str = None,
        runtime_params: Dict[str, Any] = None,
        config_patterns: Dict[str, List[str]] = None,
        *,
        base_env: str = "base",
        default_run_env: str = "local",
        globals_pattern: Optional[str] = None,
    ):
        super().__init__(
            conf_source=conf_source, env=env, runtime_params=runtime_params or {}
        )
        self.base_env = base_env
        self.default_run_env = default_run_env
        self.globals_pattern = globals_pattern

        self.config_patterns = {
            "catalog": ["catalog*", "catalog*/**", "**/catalog*"],
            "parameters": ["parameters*", "parameters*/**", "**/parameters*"],
            "credentials": ["credentials*", "credentials*/**", "**/credentials*"],
            "logging": ["logging*", "logging*/**", "**/logging*"],
        }
        self.config_patterns.update(config_patterns or {})

        self.conf_paths = _remove_duplicates(
            [
                Path(conf_source) / self.base_env,
                Path(conf_source) / (env or default_run_env),
            ]
        )

        if globals_pattern is not None:
            globals_dict = OmegaConf.to_container(
                OmegaConf.merge(
                    *[
                        OmegaConf.load(filename)
                        for filename in _path_lookup(
                            Path(conf_source) / base_env, [globals_pattern]
                        )
                    ]
                )
            )
            self.runtime_params.update({"globals": globals_dict})

    def __getitem__(self, key):
        return self.get(*self.config_patterns[key])

    def __repr__(self):  # pragma: no cover
        return (
            f"ConfigLoader(conf_source={self.conf_source}, env={self.env}, "
            f"config_patterns={self.config_patterns}), "
            f"globals_pattern='{self.globals_pattern}')"
        )

    def _load_config_file(self, config_file: Path) -> Dict[str, Any]:
        """Load an individual config file.

        Args:
            config_file (Path): Path to a config file to process.

        Raises:
            BadConfigException: If configuration is poorly formatted and
                cannot be loaded.

        Returns:
            dict: Parsed configuration as a dictionary.
        """
        try:
            conf = OmegaConf.create(
                jinja2.Template(config_file.read_text(), autoescape=True).render(
                    app=self.runtime_params.get("app", {}),
                    env=self.runtime_params.get("env", {}),
                    globals=self.runtime_params.get("globals", {}),
                )
            )
        except Exception as exc:
            raise BadConfigException(
                f"Couldn't load config file: {config_file}"
            ) from exc
        else:
            keys_to_drop = []
            for key in conf:
                if key.startswith("_"):
                    keys_to_drop.append(key)
            for key in keys_to_drop:
                del conf[key]
            return conf

    def _load_configs(self, config_filepaths: List[Path]) -> Dict[str, Any]:
        """Recursively load all configuration files, which satisfy
        a given list of glob patterns from a specific path.
        """
        aggregate_config = OmegaConf.create()
        seen_file_to_keys = {}

        for config_filepath in config_filepaths:
            single_config = self._load_config_file(config_filepath)
            _check_duplicate_keys(seen_file_to_keys, config_filepath, single_config)
            seen_file_to_keys[config_filepath] = single_config.keys()
            aggregate_config = OmegaConf.merge(aggregate_config, single_config)

        return aggregate_config

    def _get_config_from_patterns(
        self,
        patterns: Iterable[str] = None,
    ) -> Dict[str, Any]:

        if not patterns:
            raise ValueError(
                "`patterns` must contain at least one glob pattern to match filenames against."
            )

        config = OmegaConf.create()
        processed_files = set()  # type: Set[Path]

        for conf_path in self.conf_paths:
            if not conf_path.is_dir():
                raise ValueError(
                    f"Given configuration path either does not exist "
                    f"or is not a valid directory: {str(conf_path)}"
                )

            config_filepaths = _lookup_config_filepaths(
                conf_path, patterns, processed_files
            )

            new_conf = self._load_configs(config_filepaths)

            common_keys = config.keys() & new_conf.keys()
            if common_keys:
                sorted_keys = ", ".join(sorted(common_keys))
                msg = (
                    "Config from path `%s` will override the following "
                    "existing top-level config keys: %s"
                )
                logger.info(msg, conf_path, sorted_keys)

            config = OmegaConf.merge(config, new_conf)
            processed_files |= set(config_filepaths)

        if not processed_files:
            raise MissingConfigException(
                f"No files found in {self.conf_paths} matching the glob "
                f"pattern(s): {patterns}"
            )

        return OmegaConf.to_container(config, resolve=True)

    def get(self, *patterns: str) -> Dict[str, Any]:
        """Recursively scan for configuration files, load and merge them, and
        return them in the form of a config dictionary.

        Args:
            *patterns: Glob patterns to match. Files with names matching any of
                the specified patterns will be processed.

        Returns:
            Dict[str, Any]:  A dict with the combined configuration from all configuration files.

        Note:
            Any keys that start with `_` will be ignored.
        """
        return self._get_config_from_patterns(patterns=list(patterns))



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import warnings
from typing import List, Union

warnings.filterwarnings("ignore", category=DeprecationWarning)

import numpy as np
import pandas as pd
from catboost import CatBoostRegressor
from sklearn.model_selection import train_test_split

from eensight.utils import get_categorical_cols

BOOST_PARAMS = {
    "loss_function": "RMSE",
    "iterations": 10000,
    "depth": 3,
    "bootstrap_type": "Bayesian",
    "early_stopping_rounds": 100,
    "task_type": "CPU",
    "has_time": True,
    "allow_writing_files": False,
    "verbose": False,
}


def train_boost(
    X: pd.DataFrame,
    y: pd.DataFrame,
    cat_features: Union[str, List[str]] = None,
    validation_size: float = 0.2,
    return_scores: bool = False,
):
    """Train a CatBoost regressor model.

    Args:
        X (pandas.DataFrame): Feature data.
        y (pandas.DataFrame): Target data.
        cat_features (str or list of str, optional): The name(s) of the categorical features
            in the input data. Defaults to None.
        validation_size (float, optional): The size of the validation dataset (as
            percentage of all training data) to use for identifying the optimal number
            of iterations for the Gradient Boosting model. Defaults to 0.2.
        return_scores (bool, optional): If `True`, the function will also return (as second
            result) a dictionary with the regression metrics for data used for training (`learn` key)
            and for validation (`validation` key). Defaults to False.

     Returns:
        catboost.CatBoostRegressor: The trained regressor model.

    """

    if cat_features is None:
        cat_features = get_categorical_cols(X, int_is_categorical=False)

    X_train, X_val, y_train, y_val = train_test_split(
        X,
        y,
        test_size=validation_size,
        shuffle=False,
    )

    model = CatBoostRegressor(**BOOST_PARAMS).fit(
        X_train,
        y_train,
        eval_set=(X_val, y_val),
        cat_features=cat_features,
        use_best_model=True,
        verbose=False,
    )

    scores = model.get_best_score()
    scores["validation"]["CVRMSE"] = (
        scores["validation"]["RMSE"] / np.array(y_val).mean()
    )

    iterations = model.get_best_iteration() if model.get_best_iteration() > 100 else 100
    model = CatBoostRegressor(**dict(BOOST_PARAMS, iterations=iterations)).fit(
        X,
        y,
        cat_features=cat_features,
        verbose=False,
    )

    scores["learn"] = model.get_best_score()["learn"]
    scores["learn"]["CVRMSE"] = scores["learn"]["RMSE"] / np.array(y).mean()

    if return_scores:
        return model, scores

    return model



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

from kedro.pipeline import Pipeline, node, pipeline

from .nodes import evaluate_inputs, validate_inputs

preprocess_ppl = Pipeline(
    [
        node(
            func=validate_inputs,
            inputs=[
                "input-features",
                "input-labels",
                "params:rebind_names",
                "params:date_format",
                "params:validation.threshold",
                "params:alignment.mode",
                "params:alignment.tolerance",
                "params:alignment.cumulative",
            ],
            outputs=["validated-features", "validated-labels"],
            name="validate_inputs",
        ),
        node(
            func=evaluate_inputs,
            inputs=[
                "validated-features",
                "validated-labels",
                "params:filter.min_value",
                "params:filter.max_value",
                "params:filter.allow_zero",
                "params:filter.allow_negative",
                "params:adequacy.max_missing_pct",
            ],
            outputs=[
                "preprocessed-features",
                "preprocessed-labels",
                "adequacy-summary",
            ],
            name="evaluate_inputs",
        ),
    ]
)

preprocess_train = pipeline(
    preprocess_ppl,
    parameters=[
        "rebind_names",
        "date_format",
        "validation.threshold",
        "alignment.mode",
        "alignment.tolerance",
        "alignment.cumulative",
        "filter.min_value",
        "filter.max_value",
        "filter.allow_zero",
        "filter.allow_negative",
        "adequacy.max_missing_pct",
    ],
    namespace="train",
)

preprocess_test = pipeline(
    pipeline(
        preprocess_ppl.only_nodes("validate_inputs"),
        outputs={
            "validated-features": "preprocessed-features",
            "validated-labels": "preprocessed-labels",
        },
    ),
    parameters=[
        "rebind_names",
        "date_format",
        "validation.threshold",
        "alignment.mode",
        "alignment.tolerance",
        "alignment.cumulative",
    ],
    namespace="test",
)


preprocess_apply = pipeline(
    pipeline(
        preprocess_ppl.only_nodes("validate_inputs"),
        outputs={
            "validated-features": "preprocessed-features",
            "validated-labels": "preprocessed-labels",
        },
    ),
    parameters=[
        "rebind_names",
        "date_format",
        "validation.threshold",
        "alignment.mode",
        "alignment.tolerance",
        "alignment.cumulative",
    ],
    namespace="apply",
)


def create_pipeline(**kwargs):
    return (
        Pipeline([preprocess_train], tags=["train"])
        + Pipeline([preprocess_test], tags=["test"])
        + Pipeline([preprocess_apply], tags=["apply"])
    )



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Callable, Dict, Tuple

import pandas as pd

from eensight.methods.preprocessing.adequacy import (
    check_data_adequacy,
    expand_dates,
    filter_data,
)
from eensight.methods.preprocessing.alignment import validate_partitions
from eensight.utils import get_categorical_cols

logger = logging.getLogger("eensight")


############################################################################################
# Input data validation
############################################################################################


def validate_inputs(
    features: Dict[str, Callable],
    labels: Dict[str, Callable],
    rebind_names: Dict[str, str],
    date_format: str,
    threshold: float,
    mode: dict,
    tolerance: dict,
    cumulative: dict,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Validate input data.

    Args:
        features (dictionary of `load` functions): The input feature data as a partitioned
            dataset.
        labels (dictionary of `load` functions): The input label data as a partitioned
            dataset.
        rebind_names (dict): Dictionary of key/value pairs used to redefine column names
            for the input data.
        date_format (str): The strftime to parse the timestamp dates.
        threshold (float): If the range of the values that share a timestamp is less than
            `threshold` times the standard deviation of the data, they are replaced by their
            average. Otherwise, they are treated as missing values.
        mode (dict): Dictionary that maps feature names to how index alignment should be carried
            out for them. 'value' interpolates the data to match the labels' index, while 'distance'
            matches the labels' index on the nearest key of the feature's index.
        tolerance (dict): Dictionary that maps feature names to: (a) how far interpolation can reach
            if the corresponing mode is 'value' or (b) the maximum time distance to match a timestamp
            of the features dataframes with a timestamp in the labels' index if the corresponing mode
            is 'distance'.
        cumulative (dict): Dictionary that maps feature names to boolean values indicating whether the
            corresponding data is cumulative or not.

    Returns:
        tuple containing

        - **validated_features** (*pandas.DataFrame*): The validated feature data.
        - **validated_labels** (*pandas.DataFrame*): The validated label data.
    """

    tolerance = {key: pd.Timedelta(minutes=val) for key, val in tolerance.items()}

    # start with the labels
    validated_labels = validate_partitions(
        labels,
        rebind_names=rebind_names,
        date_format=date_format,
        threshold=threshold,
        mode=mode,
        tolerance=tolerance,
        cumulative=cumulative,
    )

    if len(validated_labels.columns) > 1:
        if "consumption" in validated_labels.columns:
            validated_labels = validated_labels[["consumption"]]
        else:
            raise ValueError(
                "The `labels` dataframe must contain a column named `consumption`. "
                f"Column names found: {validated_labels.columns.to_list()}"
            )

    # proceed with the features
    validated_features = validate_partitions(
        features,
        rebind_names=rebind_names,
        date_format=date_format,
        threshold=threshold,
        primary=validated_labels.index,
        mode=mode,
        tolerance=tolerance,
        cumulative=cumulative,
    )

    # fill in missing values for categorical features
    for col in get_categorical_cols(validated_features):
        validated_features[col] = validated_features[col].fillna(value="_novalue_")

    validated_labels = validated_labels.dropna()
    validated_features = validated_features.loc[
        validated_features.index.isin(validated_labels.index)
    ]
    return validated_features, validated_labels


############################################################################################
# Data adequacy
############################################################################################


def evaluate_inputs(
    features: pd.DataFrame,
    labels: pd.DataFrame,
    min_value: float,
    max_value: float,
    allow_zero: bool,
    allow_negative: bool,
    max_missing_pct: float,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Evaluate data adequacy.

    Args:
        features (pd.DataFrame): The input feature data.
        labels (pd.DataFrame): The input label data.
        min_value (float): Minumun acceptable value. Values lower than `min_value` will
            be replaced by NaN.
        max_value (float): Maximum acceptable value. Values greater than `max_value`
            will be replaced by NaN.
        allow_zero (bool): Whether zero values should be allowed.
        allow_negative (bool): Whether negative values should be allowed.
        max_missing_pct (float): The maximum acceptable percentage of missing data per month.

    Returns:
        tuple containing

        - **features** (*pandas.DataFrame*): The preprocessed feature data.
        - **labels** (*pandas.DataFrame*): The preprocessed label data.
        - **missing** (*pandas.DataFrame*): Dataframe with percentage of missing data per month.
    """

    features = expand_dates(features)
    labels = expand_dates(labels)

    labels["consumption"] = filter_data(
        labels["consumption"],
        min_value=min_value,
        max_value=max_value,
        allow_zero=allow_zero,
        allow_negative=allow_negative,
    )

    missing = check_data_adequacy(features, labels, max_missing_pct)

    labels = labels.dropna()
    features = features.loc[features.index.isin(labels.index)]
    return features, labels, missing



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

from kedro.pipeline import Pipeline, node, pipeline

from .nodes import (
    evaluate_on_apply,
    evaluate_on_apply_autoenc,
    evaluate_on_test,
    evaluate_on_test_autoenc,
)

ppl_nodes = {
    "evaluate_on_test": node(
        func=evaluate_on_test,
        inputs=["preprocessed-labels", "prediction"],
        outputs="performance",
        name="evaluate_on_test",
    ),
    "evaluate_on_test_autoenc": node(
        func=evaluate_on_test_autoenc,
        inputs=[
            "preprocessed-labels",
            "prediction-autoenc",
            "activity",
            "params:activity.assume_hurdle",
        ],
        outputs="performance-autoenc",
        name="evaluate_on_test_autoenc",
    ),
    "evaluate_on_apply": node(
        func=evaluate_on_apply,
        inputs=["preprocessed-labels", "prediction"],
        outputs="impact",
        name="evaluate_on_apply",
    ),
    "evaluate_on_apply_autoenc": node(
        func=evaluate_on_apply_autoenc,
        inputs=[
            "preprocessed-labels",
            "prediction-autoenc",
            "activity",
            "params:activity.assume_hurdle",
        ],
        outputs="impact-autoenc",
        name="evaluate_on_apply_autoenc",
    ),
}


evaluate_test = pipeline(
    pipeline([ppl_nodes["evaluate_on_test"]]),
    namespace="test",
)


evaluate_test_autoenc = pipeline(
    pipeline([ppl_nodes["evaluate_on_test_autoenc"]]),
    parameters=["activity.assume_hurdle"],
    namespace="test",
)

evaluate_apply = pipeline(
    pipeline([ppl_nodes["evaluate_on_apply"]]),
    namespace="apply",
)

evaluate_apply_autoenc = pipeline(
    pipeline([ppl_nodes["evaluate_on_apply_autoenc"]]),
    parameters=["activity.assume_hurdle"],
    namespace="apply",
)


def create_pipeline(**kwargs):
    return (
        Pipeline([evaluate_test], tags=["test_default"])
        + Pipeline([evaluate_test_autoenc], tags=["test_autoenc"])
        + Pipeline([evaluate_apply], tags=["apply_default"])
        + Pipeline([evaluate_apply_autoenc], tags=["apply_autoenc"])
    )



import os
import re
from glob import glob
from pathlib import Path

from setuptools import find_packages, setup

name = "eensight"
entry_point = "eensight = eensight.__main__:main"

here = Path(os.path.dirname(__file__)).resolve()

# get package version
with open(os.path.join(str(here), name, "__init__.py"), encoding="utf-8") as f:
    result = re.search(r'__version__ = ["\']([^"\']+)', f.read())
    if not result:
        raise ValueError("Can't find the version in eensight/__init__.py")
    version = result.group(1)


# Get the long description from the README file
with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    readme = f.read()


# get the dependencies and installs
with open("requirements.txt", encoding="utf-8") as f:
    # Make sure we strip all comments and options (e.g "--extra-index-url")
    # that arise from a modified pip.conf file that configure global options
    # when running kedro build-reqs
    requires = []
    for line in f:
        req = line.split("#", 1)[0].strip()
        if req and not req.startswith("--"):
            requires.append(req)


data_files = [name for name in glob(str(here / name / "conf") + "/**/*.*", recursive=True)]

configuration = {
    "name": name,
    "version": version,
    "python_requires": ">=3.7",
    "description": (
        "A library for measurement and verification of energy efficiency projects."
    ),
    "long_description": readme,
    "long_description_content_type": "text/markdown",
    "classifiers": [
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
    ],
    "keywords": "measurement, verification, pipelines",
    "url": "https://github.com/hebes-io/eensight",
    "maintainer": "Sotiris Papadelis",
    "maintainer_email": "spapadel@gmail.com",
    "license": "Apache License, Version 2.0",
    "packages": find_packages(exclude=["tests"]),
    "entry_points": {"console_scripts": [entry_point]},
    "install_requires": requires,
    "include_package_data": True,
    "package_data": {name: data_files},
}

setup(**configuration)



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path
from typing import Any, Dict, Union

from kedro.config import ConfigLoader
from kedro.framework.context import KedroContext
from pluggy import PluginManager


class Context(KedroContext):
    """Kedro context object.

    Args:
        package_name (str): Package name for the Kedro project the context is
            created for.
        project_path (pathlib.Path or str): Project path to define the context for.
        config_loader (kedro.config.ConfigLoader): The loader for the project's
            configuration files.
        hook_manager: The ``PluginManager`` to activate hooks, supplied by the session.
        env: Optional argument for configuration default environment to be used
            for running the pipeline. If not specified, it defaults to "local".
        extra_params: Optional dictionary containing extra project parameters.
            If specified, will update (and therefore take precedence over)
            the parameters retrieved from the project configuration.
    """

    def __init__(
        self,
        package_name: str,
        project_path: Union[Path, str],
        config_loader: ConfigLoader,
        hook_manager: PluginManager,
        env: str = None,
        extra_params: Dict[str, Any] = None,
    ):  # pylint: disable=too-many-arguments
        super().__init__(
            package_name=package_name,
            project_path=project_path,
            config_loader=config_loader,
            hook_manager=hook_manager,
            env=env,
            extra_params=extra_params,
        )

    def _get_feed_dict(self) -> Dict[str, Any]:
        """Get parameters and return the feed dictionary."""
        params = self.params
        params.pop("globals", None)
        app_data = params.pop("app", None)
        feed_dict = {"parameters": params, "app": app_data}

        def _add_param_to_feed_dict(param_name, param_value):
            """This recursively adds parameter paths to the `feed_dict`,
            whenever `param_value` is a dictionary itself, so that users can
            specify specific nested parameters in their node inputs.
            """
            key = f"params:{param_name}"
            feed_dict[key] = param_value

            if isinstance(param_value, dict):
                for key, val in param_value.items():
                    _add_param_to_feed_dict(f"{param_name}.{key}", val)

        for param_name, param_value in params.items():
            _add_param_to_feed_dict(param_name, param_value)

        return feed_dict



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import List, Union

import pandas as pd

from eensight.methods.prediction.activity import adjust_activity
from eensight.methods.prediction.baseline import UsagePredictor

logger = logging.getLogger("eensight")


def adjust_activity_levels(
    features: pd.DataFrame,
    labels: pd.DataFrame,
    model: UsagePredictor,
    non_occ_features: Union[str, List[str]],
    cat_features: Union[str, List[str]],
    assume_hurdle: bool,
    n_trials: int,
    upper_bound: float,
    verbose: bool,
):
    """Adjust activity levels for events that affect density of energy consumption.

    Args:
        features (pd.DataFrame): The feature data.
        labels (pd.DataFrame): The target data.
        model (UsagePredictor): A trained model that predicts energy consumption given
            `non_occ_features` and activity levels.
        non_occ_features (str or list of str): The names of the occupancy-independent
            features to use for estimating activity levels.
        cat_features (str or list of str): The name(s) of the categorical features in
            the input data.
        assume_hurdle (bool): If True, it is assumed that the energy consumption data is
            generated from a hurdle model.
        n_trials (int): The number of iterations for the underlying optimization.
        upper_bound (float): The maximum value that the true upper bound for the adjusted
            activity levels can take.
        verbose (bool): If True, a progress bar is visible.

    Returns:
        pandas.DataFrame: The adjusted activity time series.
    """

    activity, best_params = adjust_activity(
        features,
        labels,
        model,
        non_occ_features=non_occ_features,
        cat_features=cat_features,
        assume_hurdle=assume_hurdle,
        n_trials=n_trials,
        upper_bound=upper_bound,
        verbose=verbose,
        return_params=True,
    )

    logger.info(f"Estimated parameters: {best_params}")
    return activity.to_frame("activity")



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

from kedro.pipeline import Pipeline, node, pipeline

from .nodes import (
    create_activity_feature,
    fit_predict,
    fit_predict_autoenc,
    predict,
    predict_autoenc,
)

ppl_nodes = {
    "create_activity_feature": node(
        func=create_activity_feature,
        inputs=[
            "preprocessed-features",
            "preprocessed-labels",
            "params:activity.non_occ_features",
            "params:activity.cat_features",
            "params:activity.exog",
            "params:activity.assume_hurdle",
            "params:activity.n_trials",
            "params:activity.verbose",
            "params:activity.adjusted_activity",
        ],
        outputs="activity",
        name="create_activity_feature",
    ),
    "fit_predict": node(
        func=fit_predict,
        inputs=[
            "preprocessed-features",
            "preprocessed-labels",
            "params:fit.lags",
            "params:fit.cat_features",
            "params:fit.validation_size",
        ],
        outputs=["model", "prediction", "performance"],
        name="fit_predict",
    ),
    "fit_predict_autoenc": node(
        func=fit_predict_autoenc,
        inputs=[
            "preprocessed-features",
            "preprocessed-labels",
            "activity",
            "params:fit.lags",
            "params:fit.cat_features",
            "params:fit.validation_size",
            "params:activity.assume_hurdle",
        ],
        outputs=["model-autoenc", "prediction-autoenc", "performance-autoenc"],
        name="fit_predict_autoenc",
    ),
    "predict": node(
        func=predict,
        inputs=["preprocessed-features", "model"],
        outputs="prediction",
        name="predict",
    ),
    "predict_autoenc": node(
        func=predict_autoenc,
        inputs=[
            "preprocessed-features",
            "activity",
            "model-autoenc",
            "params:activity.assume_hurdle",
        ],
        outputs="prediction-autoenc",
        name="predict_autoenc",
    ),
}


predict_train = pipeline(
    pipeline([ppl_nodes["fit_predict"]]),
    outputs=["model"],
    parameters=[
        "fit.lags",
        "fit.cat_features",
        "fit.validation_size",
    ],
    namespace="train",
)


predict_train_autoenc = pipeline(
    pipeline(
        [
            ppl_nodes["create_activity_feature"],
            ppl_nodes["fit_predict_autoenc"],
        ],
    ),
    outputs=["model-autoenc"],
    parameters=[
        "fit.lags",
        "fit.cat_features",
        "fit.validation_size",
        "activity.non_occ_features",
        "activity.cat_features",
        "activity.exog",
        "activity.assume_hurdle",
        "activity.n_trials",
        "activity.verbose",
        "activity.adjusted_activity",
    ],
    namespace="train",
)


predict_test = pipeline(
    pipeline([ppl_nodes["predict"]]),
    inputs=["model"],
    namespace="test",
)


predict_test_autoenc = pipeline(
    pipeline(
        [
            ppl_nodes["create_activity_feature"],
            ppl_nodes["predict_autoenc"],
        ],
    ),
    inputs=["model-autoenc"],
    parameters=[
        "activity.non_occ_features",
        "activity.cat_features",
        "activity.exog",
        "activity.assume_hurdle",
        "activity.n_trials",
        "activity.verbose",
        "activity.adjusted_activity",
    ],
    namespace="test",
)


predict_apply = pipeline(
    pipeline([ppl_nodes["predict"]]),
    inputs=["model"],
    namespace="apply",
)


predict_apply_autoenc = pipeline(
    pipeline(
        [
            ppl_nodes["create_activity_feature"],
            ppl_nodes["predict_autoenc"],
        ],
    ),
    inputs=["model-autoenc"],
    parameters=[
        "activity.non_occ_features",
        "activity.cat_features",
        "activity.exog",
        "activity.assume_hurdle",
        "activity.n_trials",
        "activity.verbose",
        "activity.adjusted_activity",
    ],
    namespace="apply",
)


def create_pipeline(**kwargs):
    return (
        Pipeline([predict_train], tags=["train_default"])
        + Pipeline([predict_test], tags=["test_default"])
        + Pipeline([predict_apply], tags=["apply_default"])
        + Pipeline([predict_train_autoenc], tags=["train_autoenc"])
        + Pipeline([predict_test_autoenc], tags=["test_autoenc"])
        + Pipeline([predict_apply_autoenc], tags=["apply_autoenc"])
    )



import importlib.resources

from kedro.io import DataCatalog

import eensight

from .config import ConfigLoader
from .framework.context import Context
from .hooks import ActivityHooks, MLflowHooks

with importlib.resources.path(eensight, "__main__.py") as main_path:
    PROJECT_PATH = main_path.resolve().parent

# Instantiated project hooks.
HOOKS = (
    ActivityHooks(),
    MLflowHooks(),
)

# Installed plugins for which to disable hook auto-registration.
# DISABLE_HOOKS_FOR_PLUGINS = ("kedro-viz",)

# Class that manages storing KedroSession data.
# from kedro.framework.session.shelvestore import ShelveStore
# SESSION_STORE_CLASS = ShelveStore
# Keyword arguments to pass to the `SESSION_STORE_CLASS` constructor.
# SESSION_STORE_ARGS = {
#     "path": "./sessions"
# }

# Class that manages Kedro's library components.
CONTEXT_CLASS = Context

# Directory that holds configuration.
CONF_SOURCE = "conf"

# Class that manages how configuration is loaded.
CONFIG_LOADER_CLASS = ConfigLoader
# Keyword arguments to pass to the `CONFIG_LOADER_CLASS` constructor.
# CONFIG_LOADER_ARGS = {
#       "config_patterns": {
#           "spark" : ["spark*/"],
#           "parameters": ["parameters*", "parameters*/**", "**/parameters*"],
#       }
# }
CONFIG_LOADER_ARGS = {
    "globals_pattern": "*globals.*",
}

# Class that manages the Data Catalog.
DATA_CATALOG_CLASS = DataCatalog
